# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module related to the hardware or environment tests."""

# Pyright does not understand AntaTest.Input typing
# pyright: reportAttributeAccessIssue=false
from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, ClassVar, Literal

from pydantic import Field

from anta.custom_types import ModuleStatus, Percent, PositiveInteger, PowerSupplyFanStatus, PowerSupplyStatus
from anta.decorators import skip_on_platforms
from anta.input_models.hardware import AdverseDropThresholds, HardwareInventory, PCIeThresholds
from anta.models import AntaCommand, AntaTemplate, AntaTest
from anta.result_manager.models import AntaTestStatus

if TYPE_CHECKING:
    from collections.abc import Callable


class VerifyTransceiversManufacturers(AntaTest):
    """Verifies if all the transceivers come from approved manufacturers.

    Expected Results
    ----------------
    * Success: The test will pass if all transceivers are from approved manufacturers.
    * Failure: The test will fail if some transceivers are from unapproved manufacturers.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyTransceiversManufacturers:
          manufacturers:
            - Not Present
            - Arista Networks
            - Arastra, Inc.
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show inventory", revision=2)]
    _atomic_support: ClassVar[bool] = True

    class Input(AntaTest.Input):
        """Input model for the VerifyTransceiversManufacturers test."""

        manufacturers: list[str]
        """List of approved transceivers manufacturers."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyTransceiversManufacturers."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output

        for port, value in command_output["xcvrSlots"].items():
            # Atomic result
            result = self.result.add(description=f"Port: {port}", status=AntaTestStatus.SUCCESS)

            if not (mfg_name := value["mfgName"]):
                # Cover transceiver issues like 'xcvr-unsupported'
                result.is_failure("Manufacturer name is not available - This may indicate an unsupported or faulty transceiver")
                continue

            if mfg_name not in self.inputs.manufacturers:
                result.is_failure(f"Transceiver is from unapproved manufacturers - Expected: {', '.join(self.inputs.manufacturers)} Actual: {mfg_name}")


class VerifyTemperature(AntaTest):
    """Verifies if the device temperature is within acceptable limits.

    Expected Results
    ----------------
    * Success: The test will pass if the system temperature is `temperatureOk` and if checked, all sensor statuses and temperatures are within operational limits.
    * Failure: The test will fail if the system temperature is not `temperatureOk` or if any checked sensor reports a hardware fault or high temperature.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyTemperature:
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show system environment temperature", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyTemperature test."""

        check_temp_sensors: bool = False
        """If True, also verifies the hardware status and temperature of individual sensors."""
        failure_margin: PositiveInteger = Field(default=5)
        """"Proactive failure margin in °C. The test will fail if the current temperature is above the overheat threshold minus this margin."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyTemperature."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output
        temperature_status = command_output.get("systemStatus", "")

        # Verify system temperature status
        if temperature_status != "temperatureOk":
            self.result.is_failure(f"Device temperature exceeds acceptable limits - Expected: temperatureOk Actual: {temperature_status}")

        # Check all sensors only if check_temp_sensors knob is set.
        if not self.inputs.check_temp_sensors:
            return

        temp_sensors = command_output["tempSensors"]
        for power_supply in command_output["powerSupplySlots"]:
            temp_sensors.extend(power_supply["tempSensors"])

        for card_slot in command_output["cardSlots"]:
            temp_sensors.extend(card_slot["tempSensors"])

        for sensor in temp_sensors:
            # Account for PhyAlaska chips that don't give current temp in 7020TR
            if "PhyAlaska" in (sensor_desc := sensor["description"]):
                self.logger.debug("Sensor: %s Description: %s has been ignored due to PhyAlaska in sensor description", sensor, sensor_desc)
                continue

            # Verify sensor hardware state
            if sensor["hwStatus"] != "ok":
                self.result.is_failure(f"Sensor: {sensor['name']} Description: {sensor_desc} - Invalid hardware status - Expected: ok Actual: {sensor['hwStatus']}")
                continue

            # Verify sensor current temperature
            overheat_threshold = sensor["overheatThreshold"]
            effective_threshold = overheat_threshold - self.inputs.failure_margin
            if (act_temp := sensor["currentTemperature"]) > effective_threshold:
                self.result.is_failure(
                    f"Sensor: {sensor['name']} Description: {sensor_desc} - Temperature is getting high - Expected: <= {effective_threshold:.2f}°C "
                    f"(Overheat: {overheat_threshold:.2f}°C - Margin: {self.inputs.failure_margin}°C) Actual: {act_temp:.2f}°C"
                )


class VerifyTransceiversTemperature(AntaTest):
    """Verifies if all the transceivers are operating at an acceptable temperature.

    Expected Results
    ----------------
    * Success: The test will pass if all transceivers status are OK: 'ok'.
    * Failure: The test will fail if some transceivers are NOT OK.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyTransceiversTemperature:
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show system environment temperature transceiver", revision=1)]

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyTransceiversTemperature."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output
        sensors = command_output.get("tempSensors", "")

        for sensor in sensors:
            if sensor["hwStatus"] != "ok":
                self.result.is_failure(f"Sensor: {sensor['name']} - Invalid hardware state - Expected: ok Actual: {sensor['hwStatus']}")
            if sensor["alertCount"] != 0:
                self.result.is_failure(f"Sensor: {sensor['name']} - Incorrect alert counter - Expected: 0 Actual: {sensor['alertCount']}")


class VerifyEnvironmentSystemCooling(AntaTest):
    """Verifies the device's system cooling status.

    Expected Results
    ----------------
    * Success: The test will pass if the system cooling status is OK: 'coolingOk'.
    * Failure: The test will fail if the system cooling status is NOT OK.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyEnvironmentSystemCooling:
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show system environment cooling", revision=1)]

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyEnvironmentSystemCooling."""
        command_output = self.instance_commands[0].json_output
        sys_status = command_output.get("systemStatus", "")
        self.result.is_success()
        if sys_status != "coolingOk":
            self.result.is_failure(f"Device system cooling status invalid - Expected: coolingOk Actual: {sys_status}")


class VerifyEnvironmentCooling(AntaTest):
    """Verifies the status of power supply fans and all fan trays.

    Expected Results
    ----------------
    * Success: The test will pass if all fans have a status within the accepted states and if a speed limit is provided,
     their speed is within that limit.
    * Failure: The test will fail if any fan status is not in the accepted states or if any fan configured speed exceeds
     the speed limit, if provided.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyEnvironmentCooling:
          states:
            - ok
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show system environment cooling", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyEnvironmentCooling test."""

        states: list[PowerSupplyFanStatus]
        """List of accepted states of fan status."""
        configured_fan_speed_limit: PositiveInteger | None = None
        """The upper limit for the configured fan speed."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyEnvironmentCooling."""
        command_output = self.instance_commands[0].json_output
        self.result.is_success()
        # First go through power supplies fans
        for power_supply in command_output.get("powerSupplySlots", []):
            for fan in power_supply.get("fans", []):
                # Verify the fan status
                if (state := fan["status"]) not in self.inputs.states:
                    self.result.is_failure(
                        f"Power Slot: {power_supply['label']} Fan: {fan['label']} - Invalid state - Expected: {', '.join(self.inputs.states)} Actual: {state}"
                    )
                # Verify the configured fan speed
                elif self.inputs.configured_fan_speed_limit and fan["configuredSpeed"] > self.inputs.configured_fan_speed_limit:
                    self.result.is_failure(
                        f"Power Slot: {power_supply['label']} Fan: {fan['label']} - High fan speed - Expected: <= {self.inputs.configured_fan_speed_limit} "
                        f"Actual: {fan['configuredSpeed']}"
                    )
        # Then go through fan trays
        for fan_tray in command_output.get("fanTraySlots", []):
            for fan in fan_tray.get("fans", []):
                # Verify the fan status
                if (state := fan["status"]) not in self.inputs.states:
                    self.result.is_failure(
                        f"Fan Tray: {fan_tray['label']} Fan: {fan['label']} - Invalid state - Expected: {', '.join(self.inputs.states)} Actual: {state}"
                    )
                # Verify the configured fan speed
                elif self.inputs.configured_fan_speed_limit and fan["configuredSpeed"] > self.inputs.configured_fan_speed_limit:
                    self.result.is_failure(
                        f"Fan Tray: {fan_tray['label']} Fan: {fan['label']} - High fan speed - Expected: <= {self.inputs.configured_fan_speed_limit} "
                        f"Actual: {fan['configuredSpeed']}"
                    )


class VerifyEnvironmentPower(AntaTest):
    """Verifies the power supplies state and input voltage.

    Expected Results
    ----------------
    * Success: The test will pass if all power supplies are in an accepted state and their input voltage is greater than or equal to `min_input_voltage`
    (if provided).
    * Failure: The test will fail if any power supply is in an unaccepted state or its input voltage is less than `min_input_voltage` (if provided).

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyEnvironmentPower:
          states:
            - ok
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show system environment power", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyEnvironmentPower test."""

        states: list[PowerSupplyStatus]
        """List of accepted states for power supplies."""
        min_input_voltage: PositiveInteger | None = None
        """Optional minimum input voltage (Volts) to verify."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyEnvironmentPower."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output
        power_supplies = command_output.get("powerSupplies", "{}")
        for power_supply, value in dict(power_supplies).items():
            if (state := value["state"]) not in self.inputs.states:
                self.result.is_failure(f"Power Slot: {power_supply} - Invalid power supplies state - Expected: {', '.join(self.inputs.states)} Actual: {state}")

            # Verify if the power supply voltage is greater than the minimum input voltage
            if self.inputs.min_input_voltage and value["inputVoltage"] < self.inputs.min_input_voltage:
                self.result.is_failure(
                    f"Power Supply: {power_supply} - Input voltage mismatch - Expected: >= {self.inputs.min_input_voltage} Actual: {value['inputVoltage']}"
                )


class VerifyAdverseDrops(AntaTest):
    """Verifies there are no adverse drops exceeding defined thresholds.

    Compatible with Arista 7280R, 7500R, 7800R and 7700R series platforms supporting hardware counters.

    !!! note
        The `ReassemblyErrors` counter on a FAP can increment as a direct symptom of `FCS errors` on one of its ingress
        interfaces. This test can be configured to not fail on `ReassemblyErrors` if this correlation is found, treating
        them as an expected side effect of the initial FCS errors.

    Expected Results
    ----------------
    * Success: The test will pass if all adverse drop counters are within their defined thresholds.
    * Failure: The test will fail if any adverse drop counter exceeds its threshold.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyAdverseDrops:
          thresholds:  # Optional
            minute: 3
            ten_minute: 20
            hour: 100
            day: 500
            week: 1000
          always_fail_on_reassembly_errors: false
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaCommand(command="show hardware counter drop rates", revision=1),
        AntaCommand(command="show platform fap mapping", revision=5),
        AntaCommand(command="show interfaces counters errors", revision=1),
    ]

    class Input(AntaTest.Input):
        """Input model for the VerifyAdverseDrops test."""

        thresholds: AdverseDropThresholds = Field(default_factory=AdverseDropThresholds)
        """Adverse drop counter thresholds."""
        always_fail_on_reassembly_errors: bool = True
        """If False, the test will not fail on `ReassemblyErrors` if the same FAP reports FCS errors on one of its interfaces."""

    def _get_faps_with_errors(self, arad_mappings_output: list[dict[str, Any]], interfaces_with_errors: set[str]) -> dict[str, set[str]]:
        """Build a mapping of FAP names to a set of their interfaces that have reported FCS errors."""
        faps_with_errors = defaultdict(set)
        for fap in arad_mappings_output:
            for port_mapping in fap["portMappings"].values():
                interface_name = port_mapping["interface"]
                if interface_name.startswith("Ethernet") and interface_name in interfaces_with_errors:
                    faps_with_errors[fap["fapName"]].add(interface_name)
        return dict(faps_with_errors)

    def _get_interfaces_with_errors(self, interface_error_counters_output: dict[str, dict[str, int]]) -> set[str]:
        """Parse interface counters to find all interfaces with non-zero FCS errors."""
        return {intf_name for intf_name, counters in interface_error_counters_output.items() if counters["fcsErrors"] > 0}

    def _verify_drop_event_thresholds(self, fap_name: str, drop_event: dict[str, Any]) -> None:
        """Check the drop event against each threshold defined in the input."""
        # Iterate over the fields of the Pydantic Input model
        for field_name, field_info in AdverseDropThresholds.model_fields.items():
            # Get the eAPI key from the Field alias (e.g., "dropInLastMinute")
            eapi_key = field_info.alias

            if eapi_key not in drop_event:
                continue

            actual_value = drop_event[eapi_key]
            threshold_value = getattr(self.inputs.thresholds, field_name)

            if actual_value > threshold_value:
                counter_name = drop_event["counterName"]
                failure_msg_prefix = f"FAP: {fap_name} Counter: {counter_name}"

                # Get the human-readable period from the Field description
                human_readable_period = field_info.description
                self.result.is_failure(
                    f"{failure_msg_prefix} - {human_readable_period} rate above threshold - Expected: <= {threshold_value} Actual: {actual_value}"
                )

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyAdverseDrops."""
        self.result.is_success()

        # Extract JSON output from the commands
        show_hardware_counter_drop_rates_output = self.instance_commands[0].json_output
        show_platform_fap_mapping_output = self.instance_commands[1].json_output
        show_interfaces_counters_errors_output = self.instance_commands[2].json_output

        # Pre-build mappings for efficient lookups later
        interfaces_with_errors = self._get_interfaces_with_errors(show_interfaces_counters_errors_output["interfaceErrorCounters"])
        faps_with_errors = self._get_faps_with_errors(show_platform_fap_mapping_output["aradMappings"], interfaces_with_errors)

        for fap_name, fap_data in show_hardware_counter_drop_rates_output["dropEvents"].items():
            for drop_event in fap_data["dropEvent"]:
                # Skip events that are not 'Adverse' or have a zero drop count, as they are not relevant
                if drop_event["counterType"] != "Adverse" or drop_event["dropCount"] == 0:
                    continue

                # Special handling for 'ReassemblyErrors': log a warning message instead of failing under specific conditions
                if drop_event["counterName"] == "ReassemblyErrors" and not self.inputs.always_fail_on_reassembly_errors and fap_name in faps_with_errors:
                    fap_interfaces_with_errors = ", ".join(sorted(faps_with_errors[fap_name]))
                    self.logger.warning(
                        "%s on %s had reassembly errors but interfaces on the same FAP had FCS errors: %s", fap_name, self.device.name, fap_interfaces_with_errors
                    )
                    continue

                # Verify each threshold
                self._verify_drop_event_thresholds(fap_name, drop_event)


class VerifySupervisorRedundancy(AntaTest):
    """Verifies the redundancy protocol configured on the active supervisor.

    Expected Results
    ----------------
    * Success: The test will pass if the expected redundancy protocol is configured and operational, and if switchover is ready.
    * Failure: The test will fail if the expected redundancy protocol is not configured, not operational, or if switchover is not ready.
    * Skipped: The test will be skipped if the peer supervisor card is not inserted.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifySupervisorRedundancy:
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show redundancy status", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifySupervisorRedundancy test."""

        redundancy_proto: Literal["sso", "rpr", "simplex"] = "sso"
        """Configured redundancy protocol."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifySupervisorRedundancy."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output

        # Verify peer supervisor card insertion
        if command_output["peerState"] == "notInserted":
            self.result.is_skipped("Peer supervisor card not inserted")
            return

        # Verify that the expected redundancy protocol is configured
        if (act_proto := command_output["configuredProtocol"]) != self.inputs.redundancy_proto:
            self.result.is_failure(f"Configured redundancy protocol mismatch - Expected {self.inputs.redundancy_proto} Actual: {act_proto}")

        # Verify that the expected redundancy protocol configured and operational
        elif (act_proto := command_output["operationalProtocol"]) != self.inputs.redundancy_proto:
            self.result.is_failure(f"Operational redundancy protocol mismatch - Expected {self.inputs.redundancy_proto} Actual: {act_proto}")

        # Verify that the expected redundancy protocol configured, operational and switchover ready
        elif not command_output["switchoverReady"]:
            self.result.is_failure(f"Redundancy protocol switchover status mismatch - Expected: True Actual: {command_output['switchoverReady']}")


class VerifyPCIeErrors(AntaTest):
    """Verifies PCIe device error counters.

    Expected Results
    ----------------
    * Success: The test will pass if the correctable, non-fatal, and fatal error counts for all PCIe devices are below their defined thresholds.
    * Failure: The test will fail if any PCIe device has a correctable, non-fatal, or fatal error count above its defined threshold.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyPCIeErrors:
          thresholds:  # Optional
            correctable_errors: 10000
            non_fatal_errors: 30
            fatal_errors: 30
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show pci", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyPCIeErrors test."""

        thresholds: PCIeThresholds = Field(default_factory=PCIeThresholds)

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyPCIeErrors."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output

        for pci_id, id_details in command_output["pciIds"].items():
            for field_name, field_info in PCIeThresholds.model_fields.items():
                actual_value = id_details[field_info.alias]
                threshold_value = getattr(self.inputs.thresholds, field_name)
                if actual_value > threshold_value:
                    self.result.is_failure(
                        f"PCI Name: {id_details['name']} PCI ID: {pci_id} - {field_info.description} above threshold - "
                        f"Expected: <= {threshold_value} Actual: {actual_value}"
                    )


class VerifyAbsenceOfLinecards(AntaTest):
    """Verifies that specific linecards are not present in the device inventory.

    This is useful for confirming that hardware has been successfully decommissioned.

    Expected Results
    ----------------
    * Success: The test will pass if all provided linecard serial numbers are found.
    * Failure: The test will fail if any of the provided linecard serial numbers are found.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyAbsenceOfLinecards:
          serial_numbers:
            - VJM24220VJ1
            - VJM24230VJ2
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show inventory", revision=2)]

    class Input(AntaTest.Input):
        """Input model for the VerifyAbsenceOfLinecards test."""

        serial_numbers: list[str]
        """A list of linecard serial numbers that should NOT be in the device."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyAbsenceOfLinecards."""
        self.result.is_success()
        inventory = self.instance_commands[0].json_output
        installed_serials = {details["serialNum"] for details in inventory["cardSlots"].values()}

        # Find which of the decommissioned cards are still present
        found_serials = set(self.inputs.serial_numbers).intersection(installed_serials)
        if found_serials:
            self.result.is_failure(f"Decommissioned linecards found in inventory: {', '.join(sorted(found_serials))}")


class VerifyChassisHealth(AntaTest):
    """Verifies the health of the hardware chassis components.

    Compatible with Arista 7280R, 7500R, 7800R and 7700R series platforms.

    Expected Results
    ----------------
    * Success:  The test will pass if all linecards and fabric cards are initialized and the number of fabric interrupts does not exceed the specified threshold.
    * Failure: The test will fail if any linecards or fabric card is not initialized, or if the count of fabric interrupts is over the threshold.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyChassisHealth:
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show platform sand health", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyChassisHealth test."""

        max_fabric_interrupts: int = 0
        """The maximum number of allowed fabric interrupts."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyChassisHealth."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output

        # Verify all line cards for initialization
        if linecards_not_initialized := command_output["linecardsNotInitialized"]:
            for card in linecards_not_initialized:
                self.result.is_failure(f"Linecard: {card} - Not initialized")

        # Verify all fabric cards for initialization
        if fabric_cards_not_initialized := command_output["fabricCardsNotInitialized"]:
            for card in fabric_cards_not_initialized:
                self.result.is_failure(f"Fabric card: {card} - Not initialized")

        # Verify fabric interrupts
        for fabric, fabric_details in command_output["fabricInterruptOccurrences"].items():
            if (interrupt_count := fabric_details["count"]) > self.inputs.max_fabric_interrupts:
                self.result.is_failure(
                    f"Fabric: {fabric} - Fabric interrupts above threshold - Expected: <= {self.inputs.max_fabric_interrupts} Actual: {interrupt_count}"
                )


class VerifyInventory(AntaTest):
    """Verifies the physical hardware inventory of the device.

    By default, this test checks that all slots for the following component types
    are populated: **power supply**, **fan tray**, **fabric card**,
    **line card** and **supervisor**.

    For more granular checks, specific `requirements` can be provided.

    Expected Results
    ----------------
    * Success: The test will pass if the device inventory meets the requirements.
    * Failure: The test will fail if any component does not meet the requirements.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyInventory:
          # Verify at least 2 power supplies are installed
          # Strictly check that all fabric card slots are filled
          # Other components types (fan trays, line cards, supervisors) are ignored
          requirements:
            power_supplies: 2
            fabric_cards: all
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show inventory", revision=2)]

    class Input(AntaTest.Input):
        """Input model for the VerifyInventory test."""

        requirements: HardwareInventory | None = None
        """Specifies the required hardware inventory. If not provided, all supported components are checked."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyInventory."""
        self.result.is_success()

        command_output = self.instance_commands[0].json_output
        inventory = self._build_inventory(command_output)

        if self.inputs.requirements is None:
            self._verify_all_slots_populated(inventory)
        else:
            self._verify_specific_requirements(inventory)

    def _verify_all_slots_populated(self, inventory: dict[str, dict[str, Any]]) -> None:
        """Verify that all available slots for all component types are populated."""
        for component_data in inventory.values():
            self._report_failures(component_data)

    def _verify_specific_requirements(self, inventory: dict[str, dict[str, Any]]) -> None:
        """Verify that the inventory meets the user-defined requirements."""
        for component_type in HardwareInventory.model_fields:
            requirement = getattr(self.inputs.requirements, component_type)

            if requirement is None:
                # Requirement for this component type is not specified so we skip it
                continue

            component_data = inventory[component_type]

            if requirement == "all":
                # All available slots for this component type must be inserted
                self._report_failures(component_data)
                continue

            if isinstance(requirement, int) and (installed_count := component_data["installed"]) < requirement:
                # Check if the number of installed units meets the minimum requirement
                self.result.is_failure(f"{component_type.replace('_', ' ').title()} - Count mismatch - Expected: >= {requirement} Actual: {installed_count}")

    def _report_failures(self, component_data: dict[str, Any]) -> None:
        """Report failures for a given component type based on its state."""
        for slot in component_data.get("not_inserted", []):
            self.result.is_failure(f"{slot} - Not inserted")
        for slot in component_data.get("unidentified", []):
            self.result.is_failure(f"{slot} - Unidentified component")

    def _build_inventory(self, inventory_data: dict[str, Any]) -> dict[str, dict[str, Any]]:
        """Build a structured dictionary of the device hardware inventory."""
        inventory: dict[str, dict[str, Any]] = {
            "power_supplies": self._create_inventory_entry(),
            "fan_trays": self._create_inventory_entry(),
            "fabric_cards": self._create_inventory_entry(),
            "supervisors": self._create_inventory_entry(),
            "line_cards": self._create_inventory_entry(),
        }

        # Update inventory for each component type
        self._update_inventory_component(
            inventory=inventory,
            slots_data=inventory_data.get("powerSupplySlots", {}),
            slot_prefix="Power Supply Slot",
            name_key="name",
            component_type="power_supplies",
        )
        self._update_inventory_component(
            inventory=inventory,
            slots_data=inventory_data.get("fanTraySlots", {}),
            slot_prefix="Fan Tray Slot",
            name_key="name",
            component_type="fan_trays",
        )
        self._update_inventory_component(
            inventory=inventory,
            slots_data=inventory_data.get("cardSlots", {}),
            slot_prefix="Card Slot",
            name_key="modelName",
            component_type_getter=self._get_card_component_type,
        )

        return inventory

    def _update_inventory_component(
        self,
        inventory: dict[str, Any],
        slots_data: dict[str, dict[str, Any]],
        slot_prefix: str,
        name_key: str,
        component_type: str | None = None,
        component_type_getter: Callable[[str], str | None] | None = None,
    ) -> None:
        """Update the main inventory dictionary for a specific component type.

        This method updates the inventory dictionary by reference.
        """
        for slot, details in slots_data.items():
            current_component_type = component_type or (component_type_getter(slot) if component_type_getter else None)
            if not current_component_type:
                continue

            inventory_entry = inventory[current_component_type]
            slot_name = f"{slot_prefix}: {slot}"
            component_name = details.get(name_key)

            if not component_name:
                inventory_entry["unidentified"].append(slot_name)
            elif "Not Inserted" in component_name:
                inventory_entry["not_inserted"].append(slot_name)
            else:
                inventory_entry["installed"] += 1

    def _create_inventory_entry(self) -> dict[str, Any]:
        """Create a standard entry for a component type in the inventory."""
        return {"installed": 0, "not_inserted": [], "unidentified": []}

    def _get_card_component_type(self, card_slot_name: str) -> str | None:
        """Get the component type of a hardware card based on its slot name."""
        if card_slot_name.startswith("Fabric"):
            return "fabric_cards"
        if card_slot_name.startswith("Supervisor"):
            return "supervisors"
        if card_slot_name.startswith("Linecard"):
            return "line_cards"
        return None


class VerifyHardwareCapacityUtilization(AntaTest):
    """Verifies hardware capacity utilization.

    !!! warning
        When `strict_mode: true`, some EOS features max out hardware tables by design, which will cause failures in this mode.

    Expected Results
    ----------------
    * Success: The test will pass if all checked hardware tables are below their defined capacity utilization thresholds.
    * Failure: The test will fail if any of the checked hardware tables are above their capacity utilization threshold.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyHardwareCapacityUtilization:
          capacity_utilization_threshold: 90
          strict_mode: true
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaCommand(command="show hardware capacity alert threshold", revision=1),
        AntaTemplate(template="show hardware capacity utilization percent exceed {capacity_utilization_threshold}", revision=1),
    ]

    class Input(AntaTest.Input):
        """Input model for the VerifyHardwareCapacityUtilization test."""

        capacity_utilization_threshold: Percent = 90
        """Fails the test if the utilization of any checked hardware table exceeds this threshold."""
        strict_mode: bool = False
        """If True, check all tables. If False (default), only check tables with a configured threshold alert."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for the capacity utilization threshold."""
        return [template.render(capacity_utilization_threshold=int(self.inputs.capacity_utilization_threshold))]

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyHardwareCapacityUtilization."""
        self.result.is_success()
        alert_output = self.instance_commands[0].json_output
        utilization_output = self.instance_commands[1].json_output

        # If "tables" is empty, no thresholds were exceeded, which is a success condition
        if not (tables_exceeding_threshold := utilization_output.get("tables")):
            return

        # In strict mode, fail for any table exceeding the threshold
        if self.inputs.strict_mode:
            for table_entry in tables_exceeding_threshold:
                self.result.is_failure(self._build_failure_message(table_entry))
            return

        # Otherwise, only fail for tables that are also configured for alerting
        alert_tables = set(alert_output["thresholds"])

        for table_entry in tables_exceeding_threshold:
            table = table_entry["table"]
            feature = table_entry["feature"]
            combined_table_name = f"{table}-{feature}" if feature else table

            if combined_table_name in alert_tables:
                self.result.is_failure(self._build_failure_message(table_entry))

    def _build_failure_message(self, entry: dict[str, Any]) -> str:
        """Build the failure message from a table entry."""
        prefix_msg = f"Table: {entry['table']}"
        if chip := entry["chip"]:
            prefix_msg += f" Chip: {chip}"
        if feature := entry["feature"]:
            prefix_msg += f" Feature: {feature}"

        return f"{prefix_msg} - Capacity above threshold - Expected: < {self.inputs.capacity_utilization_threshold}% Actual: {entry['usedPercent']}%"


class VerifyModuleStatus(AntaTest):
    """Verifies the operational status and power stability of all modules in a modular chassis.

    !!! warning
        It is **crucial** to set the `supervisor_mode` input correctly to match the hardware chassis.
        Running this test in the wrong mode will result in a failure. Inventory and catalog tags can be used
        to run the test in different modes on different hardware chassis.

    Expected Results
    ----------------
    * Success: The test will pass if:
        - A dual-supervisor system has one `active` and one `standby` supervisor.
        - A single-supervisor system has one `active` supervisor.
        - All other modules are in the expected state.
        - All module risers report stable power.
    * Failure: The test will fail if any of the above conditions are not met.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyModuleStatus:
          # To accept 'ok' or 'poweredOff' statuses for linecards
          module_statuses:
            - ok
            - poweredOff
          # To test a single-supervisor chassis
          supervisor_mode: single
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaCommand(command="show module", revision=1),
        AntaCommand(command="show module power", revision=1),
    ]

    class Input(AntaTest.Input):
        """Input model for the VerifyModuleStatus test."""

        module_statuses: list[ModuleStatus] = Field(default=["ok"])
        """List of accepted statuses for modules other than supervisors (linecards, switch cards, etc.)."""
        supervisor_mode: Literal["single", "dual"] = Field(default="dual")
        """Expected supervisor configuration."""

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyModuleStatus."""
        self.result.is_success()
        modules_details = self.instance_commands[0].json_output["modules"]
        modules_power_details = self.instance_commands[1].json_output["modules"]

        if self.inputs.supervisor_mode == "dual":
            # Dual-supervisor validation logic
            supervisor_slots = {"1", "2"}
            if not supervisor_slots.issubset(modules_details):
                self.result.is_failure("Dual-Supervisor Mode - Standby supervisor is missing")
                return

            sup_statuses = {modules_details["1"]["status"], modules_details["2"]["status"]}
            expected_statuses = {"active", "standby"}

            if sup_statuses != expected_statuses:
                self.result.is_failure(
                    f"Dual-Supervisor Mode - Incorrect statuses - Expected: {'/'.join(sorted(expected_statuses))} Actual: {'/'.join(sorted(sup_statuses))}"
                )

        else:
            # Single-supervisor validation logic
            supervisor_slots = {"1"}
            if "1" not in modules_details:
                self.result.is_failure("Single-Supervisor Mode - Active supervisor is missing")
                return

            if (sup1_status := modules_details["1"]["status"]) != "active":
                self.result.is_failure(f"Single-Supervisor Mode - Incorrect status - Expected: active Actual: {sup1_status}")

        self._check_module_cards(modules_details, supervisor_slots)
        self._check_module_power(modules_power_details)

    def _check_module_cards(self, modules_details: dict[str, Any], supervisor_slots: set[str]) -> None:
        """Validate the status of all non-supervisor modules."""
        prefix = "Single-Supervisor Mode" if self.inputs.supervisor_mode == "single" else "Dual-Supervisor Mode"
        for slot, module_data in modules_details.items():
            if slot in supervisor_slots:
                continue

            module_status = module_data["status"]
            if module_status not in self.inputs.module_statuses:
                self.result.is_failure(
                    f"{prefix} - Module: {slot} Model: {module_data['modelName']} - Invalid status - "
                    f"Expected: {', '.join(self.inputs.module_statuses)} Actual: {module_status}"
                )

    def _check_module_power(self, modules_power_details: dict[str, Any]) -> None:
        """Validate that all module risers report stable power."""
        prefix = "Single-Supervisor Mode" if self.inputs.supervisor_mode == "single" else "Dual-Supervisor Mode"
        for slot, module_data in modules_power_details.items():
            for riser_slot, riser_data in module_data["risers"].items():
                if not riser_data["powerGood"]:
                    self.result.is_failure(f"{prefix} - Module: {slot} Riser {riser_slot} - Power is not stable")



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module containing input models for services tests."""

from __future__ import annotations

from ipaddress import IPv4Address, IPv6Address
from typing import Any, Literal
from warnings import warn

from pydantic import BaseModel, ConfigDict, Field

from anta.custom_types import ErrDisableReasons


class DnsServer(BaseModel):
    """Model for a DNS server configuration."""

    model_config = ConfigDict(extra="forbid")
    server_address: IPv4Address | IPv6Address
    """The IPv4 or IPv6 address of the DNS server."""
    vrf: str = "default"
    """The VRF instance in which the DNS server resides. Defaults to 'default'."""
    priority: int = Field(ge=0, le=4)
    """The priority level of the DNS server, ranging from 0 to 4. Lower values indicate a higher priority, with 0 being the highest and 4 the lowest."""

    def __str__(self) -> str:
        """Return a human-readable string representation of the DnsServer for reporting.

        Examples
        --------
        Server 10.0.0.1 (VRF: default, Priority: 1)
        """
        return f"Server {self.server_address} VRF: {self.vrf} Priority: {self.priority}"


class ErrdisableRecovery(BaseModel):
    """Model for the error disable recovery functionality."""

    model_config = ConfigDict(extra="forbid")
    reason: ErrDisableReasons
    """Name of the error disable reason."""
    status: Literal["Enabled", "Disabled"] = "Enabled"
    """Operational status of the reason. Defaults to 'Enabled'."""
    interval: int = Field(ge=30, le=86400)
    """Timer interval of the reason in seconds."""

    def __str__(self) -> str:
        """Return a human-readable string representation of the ErrdisableRecovery for reporting.

        Examples
        --------
        Reason: acl Status: Enabled Interval: 300
        """
        return f"Reason: {self.reason} Status: {self.status} Interval: {self.interval}"


class ErrDisableReason(ErrdisableRecovery):  # pragma: no cover
    """Alias for the ErrdisableRecovery model to maintain backward compatibility.

    When initialised, it will emit a deprecation warning and call the ErrdisableRecovery model.

    TODO: Remove this class in ANTA v2.0.0.
    """

    def __init__(self, **data: Any) -> None:  # noqa: ANN401
        """Initialize the ErrdisableRecovery class, emitting a depreciation warning."""
        warn(
            message="ErrDisableReason model is deprecated and will be removed in ANTA v2.0.0. Use the ErrdisableRecovery model instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**data)



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Report management for ANTA."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from jinja2 import Template
from rich.table import Table
from typing_extensions import deprecated

from anta import RICH_COLOR_PALETTE, RICH_COLOR_THEME
from anta.result_manager.models import AtomicTestResult, TestResult
from anta.tools import convert_categories

if TYPE_CHECKING:
    import pathlib

    from anta.result_manager import ResultManager
    from anta.result_manager.models import AntaTestStatus

logger = logging.getLogger(__name__)


class ReportTable:
    """Create a `rich.Table` instance based on an `anta.result_manager.ResultManager` instance.

    Attributes
    ----------
    title
        Title used when creating the `rich.Table` instance. See `ReportTable.Title` for the default values.
    columns
        Column names used when creating the `rich.Table` instance. See `ReportTable.Columns` for the default values.
    """

    @dataclass
    class Title:
        """Titles for the table report."""

        all: str = "All tests results"
        tests: str = "Summary per test"
        device: str = "Summary per device"

    @dataclass
    class Columns:  # pylint: disable=too-many-instance-attributes
        """Column names for the table report."""

        device: str = "Device"
        test: str = "Test"
        category: str = "Category"
        status: str = "Status"
        messages: str = "Message(s)"
        description: str = "Description"
        number_of_success: str = "# of success"
        number_of_failure: str = "# of failure"
        number_of_skipped: str = "# of skipped"
        number_of_errors: str = "# of errors"
        failed_devices: str = "List of devices with failed or errored tests"
        failed_tests: str = "List of failed or errored tests"

    def __init__(self) -> None:
        """Initialize a ReportTable instance."""
        self.title = ReportTable.Title()
        self.columns = ReportTable.Columns()

    def _split_list_to_txt_list(self, usr_list: list[str], delimiter: str | None = None) -> str:
        """Split list to multi-lines string.

        Parameters
        ----------
        usr_list : list[str]
            List of string to concatenate.
        delimiter : str, optional
            A delimiter to use to start string. Defaults to None.

        Returns
        -------
        str
            Multi-lines string.

        """
        if delimiter is not None:
            return "\n".join(f"{delimiter} {line}" for line in usr_list)
        return "\n".join(f"{line}" for line in usr_list)

    @staticmethod
    def _build_table(title: str, columns: list[str]) -> Table:
        """Create an empty Rich table from a title and column names.

        All the rows in the first column are colored using RICH_COLOR_PALETTE.HEADER.

        Parameters
        ----------
        title
            Title of the table.
        columns
            List of the column names.

        Returns
        -------
            A rich `Table` instance.
        """
        table = Table(title=title, show_lines=True)
        if columns:
            table.add_column(columns[0], justify="left", style=RICH_COLOR_PALETTE.HEADER, no_wrap=True)
            for column in columns[1:]:
                table.add_column(column, justify="left")
        return table

    def _color_result(self, status: AntaTestStatus) -> str:
        """Return a colored string based on an AntaTestStatus.

        Parameters
        ----------
        status
            AntaTestStatus enum to color.

        Returns
        -------
        str
            The colored string.
        """
        color = RICH_COLOR_THEME.get(str(status), "")
        return f"[{color}]{status}" if color != "" else str(status)

    def generate(self, manager: ResultManager) -> Table:
        """Create a table report with all tests.

        Attributes used to build the table are:

            Table title: `title.all`
            Table columns:
                - `columns.category`
                - `columns.device`
                - `columns.test`
                - `columns.description`
                - `columns.status`
                - `columns.messages`

        Parameters
        ----------
        manager
            A ResultManager instance.

        Returns
        -------
            A fully populated rich `Table`.
        """
        columns = [self.columns.category, self.columns.device, self.columns.test, self.columns.description, self.columns.status, self.columns.messages]

        table = ReportTable._build_table(title=self.title.all, columns=columns)

        for result in manager.results:
            state = self._color_result(result.result)
            message = self._split_list_to_txt_list(result.messages) if len(result.messages) > 0 else ""
            categories = ", ".join(convert_categories(result.categories))
            renderables: list[str | None] = [categories, str(result.name), result.test, result.description, state, message]
            table.add_row(*renderables)
        return table

    def generate_expanded(self, manager: ResultManager) -> Table:
        """Create a table report with all tests, expanded atomic results, test descriptions.

        Attributes used to build the table are:

            Table title: `title.all`
            Table columns:
                - `columns.category`
                - `columns.device`
                - `columns.test`
                - `columns.description`
                - `columns.status`
                - `columns.messages`

        Parameters
        ----------
        manager
            A ResultManager instance.

        Returns
        -------
        Table
            A fully populated rich `Table`.
        """
        columns = [
            self.columns.category,
            self.columns.device,
            self.columns.test,
            self.columns.description,
            self.columns.status,
            self.columns.messages,
        ]

        table = ReportTable._build_table(title=self.title.all, columns=columns)

        def add_line(result: TestResult | AtomicTestResult, suffix: str | None = None) -> None:
            categories = device = test = None

            if isinstance(result, TestResult):
                categories = ", ".join(convert_categories(result.categories))
                device = str(result.name)
                test = result.test
            else:  # AtomicTestResult
                if suffix is None:  # pragma: no cover
                    # This should never happen
                    msg = "Cannot generate a report line for AtomicTestResult without a suffix"
                    raise ValueError(msg)
                test = f"{result.parent.test} {suffix}"

            state = self._color_result(result.result)
            message = self._split_list_to_txt_list(result.messages) if len(result.messages) > 0 else ""
            renderables = [categories, test, device, result.description, state, message]
            table.add_row(*renderables)

        for result in manager.results:
            add_line(result)
            for index, atomic_res in enumerate(result.atomic_results):
                add_line(atomic_res, f"{index + 1}/{len(result.atomic_results)}")
        return table

    def generate_summary_by_test(self, manager: ResultManager, *, tests: set[str] | None = None) -> Table:
        """Create a table report with results aggregated per test.

        Attributes used to build the table are:

            Table title: `title.tests`
            Table columns:
                - `columns.test`
                - `columns.number_of_success`
                - `columns.number_of_skipped`
                - `columns.number_of_failure`
                - `columns.number_of_errors`
                - `columns.failed_devices`

        Parameters
        ----------
        manager
            A ResultManager instance.
        tests
            List of test names to include. None to select all tests.

        Returns
        -------
            A fully populated rich `Table`.
        """
        columns = [
            self.columns.test,
            self.columns.number_of_success,
            self.columns.number_of_skipped,
            self.columns.number_of_failure,
            self.columns.number_of_errors,
            self.columns.failed_devices,
        ]
        table = ReportTable._build_table(title=self.title.tests, columns=columns)

        for test, stats in manager.test_stats.items():
            if tests is None or test in tests:
                table.add_row(
                    test,
                    str(stats.devices_success_count),
                    str(stats.devices_skipped_count),
                    str(stats.devices_failure_count),
                    str(stats.devices_error_count),
                    ", ".join(stats.devices_failure),
                )
        return table

    def generate_summary_by_device(
        self,
        manager: ResultManager,
        *,
        devices: set[str] | None = None,
    ) -> Table:
        """Create a table report with results aggregated per device.

        Attributes used to build the table are:

            Table title: `title.device`
            Table columns:
                - `columns.device`
                - `columns.number_of_success`
                - `columns.number_of_skipped`
                - `columns.number_of_failure`
                - `columns.number_of_errors`
                - `columns.failed_tests`

        Parameters
        ----------
        manager
            A ResultManager instance.
        devices
            List of device names to include. None to select all devices.

        Returns
        -------
            A fully populated rich `Table`.
        """
        columns = [
            self.columns.device,
            self.columns.number_of_success,
            self.columns.number_of_skipped,
            self.columns.number_of_failure,
            self.columns.number_of_errors,
            self.columns.failed_tests,
        ]
        table = ReportTable._build_table(title=self.title.device, columns=columns)
        for device, stats in manager.device_stats.items():
            if devices is None or device in devices:
                table.add_row(
                    device,
                    str(stats.tests_success_count),
                    str(stats.tests_skipped_count),
                    str(stats.tests_failure_count),
                    str(stats.tests_error_count),
                    ", ".join(stats.tests_failure),
                )
        return table

    @deprecated("This method is deprecated, use `generate` instead. This will be removed in ANTA v2.0.0.", category=DeprecationWarning)
    def report_all(self, manager: ResultManager, title: str = "All tests results") -> Table:
        """Create a table report with all tests for one or all devices.

        Create table with full output: Device | Test Name | Test Status | Message(s) | Test description | Test category

        Warnings
        --------
        * This method sets the `report.title.all` value which impacts future calls to generate_* methods.

        Parameters
        ----------
        manager
            A ResultManager instance.
        title
            Title for the report. Defaults to 'All tests results'.

        Returns
        -------
        Table
            A fully populated rich `Table`.
        """
        self.title.all = title
        return self.generate(manager)

    @deprecated("This method is deprecated, use `generate_summary_tests` instead. This will be removed in ANTA v2.0.0.", category=DeprecationWarning)
    def report_summary_tests(
        self,
        manager: ResultManager,
        tests: list[str] | None = None,
        title: str = "Summary per test",
    ) -> Table:
        """Create a table report with result aggregated per test.

        Create table with full output:
        Test Name | # of success | # of skipped | # of failure | # of errors | List of failed or error nodes

        Warnings
        --------
        * This method sets the `report.title.all` value which impacts future calls to generate_* methods.

        Parameters
        ----------
        manager
            A ResultManager instance.
        tests
            List of test names to include. None to select all tests.
        title
            Title of the report.

        Returns
        -------
        Table
            A fully populated rich `Table`.
        """
        self.title.tests = title
        return self.generate_summary_by_test(manager, tests=set(tests) if tests is not None else None)

    @deprecated("This method is deprecated, use `generate_summary_devices` instead. This will be removed in ANTA v2.0.0.", category=DeprecationWarning)
    def report_summary_devices(
        self,
        manager: ResultManager,
        devices: list[str] | None = None,
        title: str = "Summary per device",
    ) -> Table:
        """Create a table report with result aggregated per device.

        Create table with full output: Device | # of success | # of skipped | # of failure | # of errors | List of failed or error test cases

        Warnings
        --------
        * This method sets the `report.title.all` value which impacts future calls to generate_* methods.

        Parameters
        ----------
        manager
            A ResultManager instance.
        devices
            List of device names to include. None to select all devices.
        title
            Title of the report.

        Returns
        -------
        Table
            A fully populated rich `Table`.
        """
        self.title.device = title
        return self.generate_summary_by_device(manager, devices=set(devices) if devices is not None else None)


class ReportJinja:  # pylint: disable=too-few-public-methods
    """Report builder based on a Jinja2 template."""

    def __init__(self, template_path: pathlib.Path) -> None:
        """Create a ReportJinja instance."""
        if not template_path.is_file():
            msg = f"template file is not found: {template_path}"
            raise FileNotFoundError(msg)

        self.template_path = template_path

    def render(self, data: list[dict[str, Any]], *, trim_blocks: bool = True, lstrip_blocks: bool = True) -> str:
        """Build a report based on a Jinja2 template.

        Report is built based on a J2 template provided by user.
        Data structure sent to template is:

        Example
        -------
        ```
        >>> print(ResultManager.json)
        [
            {
                name: ...,
                test: ...,
                result: ...,
                messages: [...]
                categories: ...,
                description: ...,
            }
        ]
        ```

        Parameters
        ----------
        data
            List of results from `ResultManager.results`.
        trim_blocks
            enable trim_blocks for J2 rendering.
        lstrip_blocks
            enable lstrip_blocks for J2 rendering.

        Returns
        -------
        str
            Rendered template

        """
        with self.template_path.open(encoding="utf-8") as file_:
            template = Template(file_.read(), trim_blocks=trim_blocks, lstrip_blocks=lstrip_blocks)

        return template.render({"data": data})



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module containing input models for BFD tests."""

from __future__ import annotations

from ipaddress import IPv4Address, IPv6Address

from pydantic import BaseModel, ConfigDict

from anta.custom_types import BfdInterval, BfdMultiplier, BfdProtocol, Interface


class BFDPeer(BaseModel):
    """Model for a BFD peer."""

    model_config = ConfigDict(extra="forbid")
    peer_address: IPv4Address | IPv6Address
    """IPv4 or IPv6 address of the BFD peer."""
    vrf: str = "default"
    """VRF of the BFD peer."""
    interface: Interface | None = None
    """Single-hop transport interface. Use `None` for multi-hop sessions."""
    protocols: list[BfdProtocol] | None = None
    """List of protocols using BFD with this peer. Required field in the `VerifyBFDPeersRegProtocols` test."""
    tx_interval: BfdInterval | None = None
    """Operational transmit interval of the BFD session in milliseconds. Required field in the `VerifyBFDPeersIntervals` test."""
    rx_interval: BfdInterval | None = None
    """Operational minimum receive interval of the BFD session in milliseconds. Required field in the `VerifyBFDPeersIntervals` test."""
    multiplier: BfdMultiplier | None = None
    """Multiplier of the BFD session. Required field in the `VerifyBFDPeersIntervals` test."""
    detection_time: int | None = None
    """Detection time of the BFD session in milliseconds. Defines how long it takes for BFD to detect connection failure.

    Optional field in the `VerifyBFDPeersIntervals` test."""

    def __str__(self) -> str:
        """Return a human-readable string representation of the BFDPeer for reporting."""
        base = f"Peer: {self.peer_address} VRF: {self.vrf}"
        if self.interface is not None:
            base += f" Interface: {self.interface}"
        return base



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module containing input models for AVT tests."""

from __future__ import annotations

from ipaddress import IPv4Address

from pydantic import BaseModel, ConfigDict


class AVTPath(BaseModel):
    """AVT (Adaptive Virtual Topology) model representing path details and associated information."""

    model_config = ConfigDict(extra="forbid")
    vrf: str = "default"
    """VRF context. Defaults to `default`."""
    avt_name: str
    """The name of the Adaptive Virtual Topology (AVT)."""
    description: str | None = None
    """Optional metadata describing the AVT path. Used for reporting."""
    destination: IPv4Address
    """The IPv4 address of the destination peer in the AVT."""
    next_hop: IPv4Address
    """The IPv4 address of the next hop used to reach the AVT peer."""
    path_type: str | None = None
    """Specifies the type of path for the AVT. If not specified, both types 'direct' and 'multihop' are considered."""

    def __str__(self) -> str:
        """Return a human-readable string representation of the AVTPath for reporting.

        Examples
        --------
        AVT: CONTROL-PLANE-PROFILE VRF: default Destination: 10.101.255.2 Next-hop: 10.101.255.1

        """
        base_string = f"AVT: {self.avt_name}"
        if self.description:
            base_string += f" ({self.description})"
        base_string += f" VRF: {self.vrf} Destination: {self.destination} Next-hop: {self.next_hop}"
        if self.path_type:
            base_string += f" Path Type: {self.path_type}"

        return base_string



# Copyright (c) 2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module containing input models for hardware tests."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, ConfigDict, Field


class AdverseDropThresholds(BaseModel):
    """Thresholds for adverse drop counters over various time periods."""

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    minute: int = Field(default=0, alias="dropInLastMinute", description="Last minute")
    """Threshold for the last minute."""
    ten_minute: int = Field(default=0, alias="dropInLastTenMinute", description="Last 10 minutes")
    """Threshold for the last ten minutes."""
    hour: int = Field(default=0, alias="dropInLastOneHour", description="Last hour")
    """Threshold for the last hour."""
    day: int = Field(default=0, alias="dropInLastOneDay", description="Last day")
    """Threshold for the last day."""
    week: int = Field(default=0, alias="dropInLastOneWeek", description="Last week")
    """Threshold for the last week."""


class PCIeThresholds(BaseModel):
    """Thresholds for PCIe device error counters."""

    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    correctable_errors: int = Field(default=0, alias="correctableErrors", description="Correctable errors")
    """Threshold for correctable errors."""
    non_fatal_errors: int = Field(default=0, alias="nonFatalErrors", description="Non-fatal errors")
    """Threshold for non-fatal errors."""
    fatal_errors: int = Field(default=0, alias="fatalErrors", description="Fatal errors")
    """Threshold for fatal errors."""


class HardwareInventory(BaseModel):
    """Specifies the required hardware inventory.

    For each field, providing a value of `None` (or omitting it) will skip
    the check for that specific hardware component category.
    """

    model_config = ConfigDict(extra="forbid")
    power_supplies: int | Literal["all"] | None = None
    """Required power supplies. Use 'all' to enforce that all slots are filled."""
    fan_trays: int | Literal["all"] | None = None
    """Required fan trays. Use 'all' to enforce that all slots are filled."""
    fabric_cards: int | Literal["all"] | None = None
    """Required fabric cards. Use 'all' to enforce that all slots are filled."""
    line_cards: int | Literal["all"] | None = None
    """Required line cards. Use 'all' to enforce that all slots are filled."""
    supervisors: int | Literal["all"] | None = None
    """Required supervisors. Use 'all' to enforce that all slots are filled."""



# Copyright (c) 2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module containing input models for routing OSPF tests."""

from __future__ import annotations

from ipaddress import IPv4Address
from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator

from anta.custom_types import Interface


class OSPFNeighbor(BaseModel):
    """Model for an OSPF neighbor."""

    model_config = ConfigDict(extra="forbid")
    instance: int = Field(ge=1, le=65535)
    """OSPF instance."""
    vrf: str = "default"
    """VRF context of the OSPF instance."""
    ip_address: IPv4Address
    """Neighbor interface IP address."""
    local_interface: Interface
    """Local interface establishing the adjacency."""
    state: Literal["full", "2Ways"] = "full"
    """Expected adjacency state."""
    area_id: IPv4Address | Annotated[int, Field(ge=0, le=4294967295)]
    """OSPF area-id in decimal or IP address format."""

    @field_validator("area_id", mode="after")
    @classmethod
    def convert_area_id(cls, area_id: IPv4Address | int) -> IPv4Address:
        """Convert a decimal OSPF area-id to an IP address format if needed."""
        return IPv4Address(area_id) if isinstance(area_id, int) else area_id

    def __str__(self) -> str:
        """Return a human-readable string representation of the OSPFNeighbor for reporting."""
        return f"Instance: {self.instance} VRF: {self.vrf} Neighbor IP: {self.ip_address} Area: {self.area_id}"



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Utils functions to use with anta.cli.get.commands module."""

from __future__ import annotations

import functools
import importlib
import inspect
import json
import logging
import pkgutil
import re
import sys
import textwrap
from importlib import util as importlib_util
from itertools import groupby
from pathlib import Path
from sys import stdin
from typing import TYPE_CHECKING, Any, TypeVar

import click
import requests
import urllib3
import yaml
from typing_extensions import deprecated

from anta.cli.console import console
from anta.cli.utils import ExitCode
from anta.inventory import AntaInventory
from anta.inventory.models import AntaInventoryHost, AntaInventoryInput
from anta.models import AntaCommand, AntaTest

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

if TYPE_CHECKING:
    from collections.abc import Callable

    from anta.catalog import AntaCatalog

R = TypeVar("R")

logger = logging.getLogger(__name__)


def inventory_output_options(f: Callable[..., R]) -> Callable[..., R]:
    """Click common options required when an inventory is being generated."""

    @click.option(
        "--output",
        "-o",
        required=True,
        envvar="ANTA_INVENTORY",
        show_envvar=True,
        help="Path to save inventory file",
        type=click.Path(file_okay=True, dir_okay=False, exists=False, writable=True, path_type=Path),
    )
    @click.option(
        "--overwrite",
        help="Do not prompt when overriding current inventory",
        default=False,
        is_flag=True,
        show_default=True,
        required=False,
        show_envvar=True,
    )
    @click.pass_context
    @functools.wraps(f)
    def wrapper(
        ctx: click.Context,
        *,
        output: Path,
        overwrite: bool,
        **kwargs: Any,  # noqa: ANN401
    ) -> R:
        # Boolean to check if the file is empty
        output_is_not_empty = output.exists() and output.stat().st_size != 0
        # Check overwrite when file is not empty
        if not overwrite and output_is_not_empty:
            is_tty = stdin.isatty()
            if is_tty:
                # File has content and it is in an interactive TTY --> Prompt user
                _ = click.confirm(
                    f"Your destination file '{output}' is not empty, continue?",
                    abort=True,
                )
            else:
                # File has content and it is not interactive TTY nor overwrite set to True --> execution stop
                logger.critical("Conversion aborted since destination file is not empty (not running in interactive TTY)")
                ctx.exit(ExitCode.USAGE_ERROR)
        output.parent.mkdir(parents=True, exist_ok=True)
        return f(output=output, **kwargs)

    return wrapper


def get_cv_token(cvp_ip: str, cvp_username: str, cvp_password: str, *, verify_cert: bool) -> str:
    """Generate the authentication token from CloudVision using username and password.

    TODO: need to handle requests error

    Parameters
    ----------
    cvp_ip
        IP address of CloudVision.
    cvp_username
        Username to connect to CloudVision.
    cvp_password
        Password to connect to CloudVision.
    verify_cert
        Enable or disable certificate verification when connecting to CloudVision.

    Returns
    -------
    str
        The token to use in further API calls to CloudVision.

    Raises
    ------
    requests.ssl.SSLError
        If the certificate verification fails.

    """
    # use CVP REST API to generate a token
    url = f"https://{cvp_ip}/cvpservice/login/authenticate.do"
    payload = json.dumps({"userId": cvp_username, "password": cvp_password})
    headers = {"Content-Type": "application/json", "Accept": "application/json"}

    response = requests.request("POST", url, headers=headers, data=payload, verify=verify_cert, timeout=10)
    return response.json()["sessionId"]


def write_inventory_to_file(hosts: list[AntaInventoryHost], output: Path) -> None:
    """Write a file inventory from pydantic models.

    Parameters
    ----------
    hosts:
        the list of AntaInventoryHost to write to an inventory file
    output:
        the Path where the inventory should be written.

    Raises
    ------
    OSError
        When anything goes wrong while writing the file.
    """
    i = AntaInventoryInput(hosts=hosts)
    try:
        with output.open(mode="w", encoding="UTF-8") as out_fd:
            _ = out_fd.write(yaml.dump({AntaInventory.INVENTORY_ROOT_KEY: yaml.safe_load(i.yaml())}))
        logger.info("ANTA inventory file has been created: '%s'", output)
    except OSError as exc:
        msg = f"Could not write inventory to path '{output}'."
        raise OSError(msg) from exc


def create_inventory_from_cvp(inv: list[dict[str, Any]], output: Path) -> None:
    """Create an inventory file from Arista CloudVision inventory."""
    logger.debug("Received %s device(s) from CloudVision", len(inv))
    hosts: list[AntaInventoryHost] = []
    for dev in inv:
        logger.info("   * adding entry for %s", dev["hostname"])
        hosts.append(
            AntaInventoryHost(
                name=dev["hostname"],
                host=dev["ipAddress"],
                tags={dev["containerName"].lower()},
            )
        )
    write_inventory_to_file(hosts, output)


def find_ansible_group(data: dict[str, Any], group: str) -> dict[str, Any] | None:
    """Retrieve Ansible group from an input data dict."""
    for k, v in data.items():
        if isinstance(v, dict):
            if k == group and ("children" in v or "hosts" in v):
                return v
            d = find_ansible_group(v, group)
            if d is not None:
                return d
    return None


def deep_yaml_parsing(data: dict[str, Any], hosts: list[AntaInventoryHost] | None = None) -> list[AntaInventoryHost]:
    """Deep parsing of YAML file to extract hosts and associated IPs."""
    if hosts is None:
        hosts = []
    for key, value in data.items():
        if isinstance(value, dict) and "ansible_host" in value:
            logger.info("   * adding entry for %s", key)
            hosts.append(AntaInventoryHost(name=key, host=value["ansible_host"]))
        elif isinstance(value, dict):
            deep_yaml_parsing(value, hosts)
        else:
            return hosts
    return hosts


def create_inventory_from_ansible(inventory: Path, output: Path, ansible_group: str = "all") -> None:
    """Create an ANTA inventory from an Ansible inventory YAML file.

    Parameters
    ----------
    inventory
        Ansible Inventory file to read.
    output
        ANTA inventory file to generate.
    ansible_group
        Ansible group from where to extract data.

    """
    try:
        with inventory.open(encoding="utf-8") as inv:
            ansible_inventory: dict[str, Any] = yaml.safe_load(inv)
    except yaml.constructor.ConstructorError as exc:
        if exc.problem and "!vault" in exc.problem:
            msg = (
                "`anta get from-ansible` does not support inline vaulted variables, comment them out to generate your inventory. "
                "If the vaulted variable is necessary to build the inventory (e.g. `ansible_host`), it needs to be unvaulted for "
                "`from-ansible` command to work."
            )
            logger.error(msg)
        msg = f"Could not parse {inventory}."
        raise ValueError(msg) from exc
    except OSError as exc:
        msg = f"Could not parse {inventory}."
        raise ValueError(msg) from exc

    if not ansible_inventory:
        msg = f"Ansible inventory {inventory} is empty"
        raise ValueError(msg)

    filtered_ansible_inventory = find_ansible_group(ansible_inventory, ansible_group)

    if filtered_ansible_inventory is None:
        msg = f"Group {ansible_group} not found in Ansible inventory"
        raise ValueError(msg)
    ansible_hosts = deep_yaml_parsing(filtered_ansible_inventory)
    write_inventory_to_file(ansible_hosts, output)


def _explore_package(module_name: str, test_name: str | None = None, *, short: bool = False, count: bool = False) -> list[type[AntaTest]]:
    """Parse ANTA test submodules recursively and return a list of the found AntaTest.

    Parameters
    ----------
    module_name
        Name of the module to explore (e.g., 'anta.tests.routing.bgp').
    test_name
        If provided, only show tests starting with this name.
    short
        If True, only print test names without their inputs.
    count
        If True, only count the tests.

    Returns
    -------
    list[type[AntaTest]]:
        A list of the AntaTest found.
    """
    result: list[type[AntaTest]] = []
    try:
        module_spec = importlib_util.find_spec(module_name)
    except ModuleNotFoundError:
        # Relying on module_spec check below.
        module_spec = None
    except ImportError as e:
        msg = "`--module <module>` option does not support relative imports"
        raise ValueError(msg) from e

    # Giving a second chance adding CWD to PYTHONPATH
    if module_spec is None:
        try:
            logger.info("Could not find module `%s`, injecting CWD in PYTHONPATH and retrying...", module_name)
            sys.path = [str(Path.cwd()), *sys.path]
            module_spec = importlib_util.find_spec(module_name)
        except ImportError:
            module_spec = None

    if module_spec is None or module_spec.origin is None:
        msg = f"Module `{module_name}` was not found!"
        raise ValueError(msg)

    if module_spec.submodule_search_locations:
        for _, sub_module_name, ispkg in pkgutil.walk_packages(module_spec.submodule_search_locations):
            qname = f"{module_name}.{sub_module_name}"
            if ispkg:
                result.extend(_explore_package(qname, test_name=test_name, short=short, count=count))
                continue
            result.extend(find_tests_in_module(qname, test_name))

    else:
        result.extend(find_tests_in_module(module_spec.name, test_name))

    return result


def find_tests_in_module(qname: str, test_name: str | None) -> list[type[AntaTest]]:
    """Return the list of AntaTest in the passed module qname, potentially filtering on test_name.

    Parameters
    ----------
    qname
        Name of the module to explore (e.g., 'anta.tests.routing.bgp').
    test_name
        If provided, only show tests starting with this name.

    Returns
    -------
    list[type[AntaTest]]:
        A list of the AntaTest found in the module.
    """
    results: list[type[AntaTest]] = []
    try:
        qname_module = importlib.import_module(qname)
    except (AssertionError, ImportError) as e:
        msg = f"Error when importing `{qname}` using importlib!"
        raise ValueError(msg) from e

    for _name, obj in inspect.getmembers(qname_module):
        # Only retrieves the subclasses of AntaTest
        if not inspect.isclass(obj) or not issubclass(obj, AntaTest) or obj == AntaTest:
            continue
        if test_name and not obj.name.startswith(test_name):
            continue
        results.append(obj)

    return results


def _filter_tests_via_catalog(tests: list[type[AntaTest]], catalog: AntaCatalog) -> list[type[AntaTest]]:
    """Return the filtered list of tests present in the catalog.

    Parameters
    ----------
    tests:
        List of tests.
    catalog:
        The AntaCatalog to use as filtering.

    Returns
    -------
    list[type[AntaTest]]:
        The filtered list of tests containing uniquely the tests found in the catalog.
    """
    catalog_test_names = {test.test.name for test in catalog.tests}
    return [test for test in tests if test.name in catalog_test_names]


def print_tests(tests: list[type[AntaTest]], *, short: bool = False) -> None:
    """Print a list of AntaTest.

    Parameters
    ----------
    tests
        A list of AntaTest subclasses.
    short
        If True, only print test names without their inputs.
    """

    def module_name(test: type[AntaTest]) -> str:
        """Return the module name for the input test.

        Used to group the test by module.
        """
        return test.__module__

    for module, module_tests in groupby(tests, module_name):
        console.print(f"{module}:")
        for test in module_tests:
            print_test(test, short=short)


def print_test(test: type[AntaTest], *, short: bool = False) -> None:
    """Print a single test.

    Parameters
    ----------
    test
        the representation of the AntaTest as returned by inspect.getmembers
    short
        If True, only print test names without their inputs.
    """
    if not test.__doc__ or (example := extract_examples(test.__doc__)) is None:
        msg = f"Test {test.name} in module {test.__module__} is missing an Example"
        raise LookupError(msg)
    # Picking up only the inputs in the examples
    # Need to handle the fact that we nest the routing modules in Examples.
    # This is a bit fragile.
    inputs = example.split("\n")
    test_name_lines = [i for i, input_entry in enumerate(inputs) if test.name in input_entry]
    if not test_name_lines:
        msg = f"Could not find the name of the test '{test.name}' in the Example section in the docstring."
        raise ValueError(msg)
    for list_index, line_index in enumerate(test_name_lines):
        end = test_name_lines[list_index + 1] if list_index + 1 < len(test_name_lines) else -1
        console.print(f"  {inputs[line_index].strip()}")
        # Injecting the description for the first example
        if list_index == 0:
            console.print(f"      # {test.description}", soft_wrap=True)
        if not short and len(inputs) > line_index + 2:  # There are params
            console.print(textwrap.indent(textwrap.dedent("\n".join(inputs[line_index + 1 : end])), " " * 6))


def extract_examples(docstring: str) -> str | None:
    """Extract the content of the Example section in a Numpy docstring.

    Returns
    -------
    str | None
        The content of the section if present, None if the section is absent or empty.
    """
    pattern = r"Examples\s*--------\s*(.*)(?:\n\s*\n|\Z)"
    match = re.search(pattern, docstring, flags=re.DOTALL)
    return match[1].strip() if match and match[1].strip() != "" else None


def _print_commands(tests: list[type[AntaTest]]) -> None:
    """Print a list of commands per module and per test.

    Parameters
    ----------
    tests
        A list of AntaTest subclasses.
    """

    def module_name(test: type[AntaTest]) -> str:
        """Return the module name for the input test.

        Used to group the test by module.
        """
        return test.__module__

    for module, module_tests in groupby(tests, module_name):
        console.print(f"{module}:")
        for test in module_tests:
            console.print(f"  - {test.name}:")
            for command in test.commands:
                if isinstance(command, AntaCommand):
                    console.print(f"    - {command.command}")
                else:  # isinstance(command, AntaTemplate):
                    console.print(f"    - {command.template}")


def _get_unique_commands(tests: list[type[AntaTest]]) -> set[str]:
    """Return a set of unique commands used by the tests.

    Parameters
    ----------
    tests
        A list of AntaTest subclasses.

    Returns
    -------
    set[str]
        A set of commands or templates used by each test.
    """
    result: set[str] = set()

    for test in tests:
        for command in test.commands:
            if isinstance(command, AntaCommand):
                result.add(command.command)
            else:  # isinstance(command, AntaTemplate):
                result.add(command.template)

    return result


@deprecated("This function is deprecated, use `_explore_package`. This will be removed in ANTA v2.0.0.", category=DeprecationWarning)
def explore_package(module_name: str, test_name: str | None = None, *, short: bool = False, count: bool = False) -> int:  # pragma: no cover
    """Parse ANTA test submodules recursively and print AntaTest examples.

    Parameters
    ----------
    module_name
        Name of the module to explore (e.g., 'anta.tests.routing.bgp').
    test_name
        If provided, only show tests starting with this name.
    short
        If True, only print test names without their inputs.
    count
        If True, only count the tests.

    Returns
    -------
    int:
        The number of tests found.
    """
    try:
        module_spec = importlib_util.find_spec(module_name)
    except ModuleNotFoundError:
        # Relying on module_spec check below.
        module_spec = None
    except ImportError as e:
        msg = "`anta get tests --module <module>` does not support relative imports"
        raise ValueError(msg) from e

    # Giving a second chance adding CWD to PYTHONPATH
    if module_spec is None:
        try:
            logger.info("Could not find module `%s`, injecting CWD in PYTHONPATH and retrying...", module_name)
            sys.path = [str(Path.cwd()), *sys.path]
            module_spec = importlib_util.find_spec(module_name)
        except ImportError:
            module_spec = None

    if module_spec is None or module_spec.origin is None:
        msg = f"Module `{module_name}` was not found!"
        raise ValueError(msg)

    tests_found = 0
    if module_spec.submodule_search_locations:
        for _, sub_module_name, ispkg in pkgutil.walk_packages(module_spec.submodule_search_locations):
            qname = f"{module_name}.{sub_module_name}"
            if ispkg:
                tests_found += explore_package(qname, test_name=test_name, short=short, count=count)
                continue
            tests_found += find_tests_examples(qname, test_name, short=short, count=count)

    else:
        tests_found += find_tests_examples(module_spec.name, test_name, short=short, count=count)

    return tests_found


@deprecated("This function is deprecated, use `find_tests_in_module`. This will be removed in ANTA v2.0.0.", category=DeprecationWarning)
def find_tests_examples(qname: str, test_name: str | None, *, short: bool = False, count: bool = False) -> int:  # pragma: no cover
    """Print tests from `qname`, filtered by `test_name` if provided.

    Parameters
    ----------
    qname
        Name of the module to explore (e.g., 'anta.tests.routing.bgp').
    test_name
        If provided, only show tests starting with this name.
    short
        If True, only print test names without their inputs.
    count
        If True, only count the tests.

    Returns
    -------
    int:
        The number of tests found.
    """
    try:
        qname_module = importlib.import_module(qname)
    except (AssertionError, ImportError) as e:
        msg = f"Error when importing `{qname}` using importlib!"
        raise ValueError(msg) from e

    module_printed = False
    tests_found = 0

    for _name, obj in inspect.getmembers(qname_module):
        # Only retrieves the subclasses of AntaTest
        if not inspect.isclass(obj) or not issubclass(obj, AntaTest) or obj == AntaTest:
            continue
        if test_name and not obj.name.startswith(test_name):
            continue
        if not module_printed:
            if not count:
                console.print(f"{qname}:")
            module_printed = True
        tests_found += 1
        if count:
            continue
        print_test(obj, short=short)

    return tests_found



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module related to system-level features and protocols tests."""

# Pyright does not understand AntaTest.Input typing
# pyright: reportAttributeAccessIssue=false
from __future__ import annotations

import re
from json import JSONDecodeError, loads
from typing import TYPE_CHECKING, Any, ClassVar, Literal

from pydantic import Field, model_validator

from anta.custom_types import Hostname, Percent, PositiveInteger, ReloadCause
from anta.decorators import skip_on_platforms
from anta.input_models.system import NTPPool, NTPServer
from anta.models import AntaCommand, AntaTemplate, AntaTest
from anta.tools import get_item, get_value

if TYPE_CHECKING:
    import sys
    from ipaddress import IPv4Address

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

CPU_IDLE_THRESHOLD = 25
MEMORY_THRESHOLD = 0.25
DISK_SPACE_THRESHOLD = 75


class VerifyUptime(AntaTest):
    """Verifies the device uptime.

    Expected Results
    ----------------
    * Success: The test will pass if the device uptime is higher than the provided value.
    * Failure: The test will fail if the device uptime is lower than the provided value.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyUptime:
          minimum: 86400
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show uptime", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyUptime test."""

        minimum: PositiveInteger
        """Minimum uptime in seconds."""

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyUptime."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output
        if command_output["upTime"] < self.inputs.minimum:
            self.result.is_failure(f"Device uptime is incorrect - Expected: {self.inputs.minimum}s Actual: {command_output['upTime']}s")


class VerifyReloadCause(AntaTest):
    """Verifies the last reload cause of the device.

    User Input to Reload Cause Mapping:

    | Input Value   | Mapped To                                     |
    |---------------|-----------------------------------------------|
    | USER          | Reload requested by the user.                 |
    | FPGA          | Reload requested after FPGA upgrade           |
    | ZTP           | System reloaded due to Zero Touch Provisioning|
    | USER_HITLESS  | Hitless reload requested by the user.         |

    Expected Results
    ----------------
    * Success: The test passes if there is no reload cause, or if the last reload cause was one of the provided inputs.
    * Failure: The test will fail if the last reload cause was NOT one of the provided inputs.
    * Error: The test will report an error if the reload cause is NOT available.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyReloadCause:
        allowed_causes:
          - USER
          - FPGA
          - ZTP
          - USER_HITLESS
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show reload cause", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyReloadCause test."""

        allowed_causes: list[ReloadCause] = Field(default=["USER", "FPGA"], validate_default=True)
        """A list of allowed system reload causes."""

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyReloadCause."""
        command_output = self.instance_commands[0].json_output
        if len(command_output["resetCauses"]) == 0:
            # No reload causes
            self.result.is_success()
            return

        reset_causes = command_output["resetCauses"]
        command_output_data = reset_causes[0].get("description")
        if command_output_data in self.inputs.allowed_causes:
            self.result.is_success()
        else:
            causes = ", ".join(f"'{c}'" for c in self.inputs.allowed_causes)
            self.result.is_failure(f"Invalid reload cause -  Expected: {causes} Actual: '{command_output_data}'")


class VerifyCoredump(AntaTest):
    """Verifies there are no core dump files.

    Expected Results
    ----------------
    * Success: The test will pass if there are NO core dump(s) in /var/core.
    * Failure: The test will fail if there are core dump(s) in /var/core.

    Notes
    -----
    * This test will NOT check for minidump(s) generated by certain agents in /var/core/minidump.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyCoredump:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show system coredump", revision=1)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyCoredump."""
        command_output = self.instance_commands[0].json_output
        core_files = command_output["coreFiles"]
        if "minidump" in core_files:
            core_files.remove("minidump")
        if not core_files:
            self.result.is_success()
        else:
            self.result.is_failure(f"Core dump(s) have been found: {', '.join(core_files)}")


class VerifyAgentLogs(AntaTest):
    """Verifies there are no agent crash reports.

    Expected Results
    ----------------
    * Success: The test will pass if there is NO agent crash reported.
    * Failure: The test will fail if any agent crashes are reported.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyAgentLogs:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show agent logs crash", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyAgentLogs."""
        command_output = self.instance_commands[0].text_output
        if len(command_output) == 0:
            self.result.is_success()
        else:
            pattern = re.compile(r"^===> (.*?) <===$", re.MULTILINE)
            agents = "\n * ".join(pattern.findall(command_output))
            self.result.is_failure(f"Device has reported agent crashes:\n * {agents}")


class VerifyCPUUtilization(AntaTest):
    """Verifies whether the CPU utilization is below 75%.

    Expected Results
    ----------------
    * Success: The test will pass if the CPU utilization is below 75%.
    * Failure: The test will fail if the CPU utilization is over 75%.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyCPUUtilization:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show processes top once", revision=1)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyCPUUtilization."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output
        command_output_data = command_output["cpuInfo"]["%Cpu(s)"]["idle"]
        if command_output_data < CPU_IDLE_THRESHOLD:
            self.result.is_failure(f"Device has reported a high CPU utilization -  Expected: < 75% Actual: {100 - command_output_data}%")


class VerifyMemoryUtilization(AntaTest):
    """Verifies whether the memory utilization is below 75%.

    Expected Results
    ----------------
    * Success: The test will pass if the memory utilization is below 75%.
    * Failure: The test will fail if the memory utilization is over 75%.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyMemoryUtilization:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show version", revision=1)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyMemoryUtilization."""
        self.result.is_success()
        command_output = self.instance_commands[0].json_output
        memory_usage = command_output["memFree"] / command_output["memTotal"]
        if memory_usage < MEMORY_THRESHOLD:
            self.result.is_failure(f"Device has reported a high memory usage - Expected: < 75% Actual: {(1 - memory_usage) * 100:.2f}%")


class VerifyFileSystemUtilization(AntaTest):
    """Verifies that no partition is utilizing more than 75% of its disk space.

    Expected Results
    ----------------
    * Success: The test will pass if all partitions are using less than 75% of its disk space.
    * Failure: The test will fail if any partitions are using more than 75% of its disk space.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyFileSystemUtilization:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="bash timeout 10 df -h", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyFileSystemUtilization."""
        command_output = self.instance_commands[0].text_output
        self.result.is_success()
        for line in command_output.split("\n")[1:]:
            if "loop" not in line and len(line) > 0 and (percentage := int(line.split()[4].replace("%", ""))) > DISK_SPACE_THRESHOLD:
                self.result.is_failure(f"Mount point: {line} - Higher disk space utilization - Expected: {DISK_SPACE_THRESHOLD}% Actual: {percentage}%")


class VerifyNTP(AntaTest):
    """Verifies if NTP is synchronised.

    Expected Results
    ----------------
    * Success: The test will pass if the NTP is synchronised.
    * Failure: The test will fail if the NTP is NOT synchronised.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyNTP:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show ntp status", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyNTP."""
        command_output = self.instance_commands[0].text_output
        if command_output.split("\n")[0].split(" ")[0] == "synchronised":
            self.result.is_success()
        else:
            data = command_output.split("\n")[0]
            self.result.is_failure(f"NTP status mismatch - Expected: synchronised Actual: {data}")


class VerifyNTPAssociations(AntaTest):
    """Verifies the Network Time Protocol (NTP) associations.

    This test performs the following checks:

      1. For the NTP servers:
        - The primary NTP server (marked as preferred) has the condition 'sys.peer'.
        - All other NTP servers have the condition 'candidate'.
        - All the NTP servers have the expected stratum level.
      2. For the NTP servers pool:
        - All the NTP servers belong to the specified NTP pool.
        - All the NTP servers have valid condition (sys.peer | candidate).
        - All the NTP servers have the stratum level within the specified startum level.

    Expected Results
    ----------------
    * Success: The test will pass if all the NTP servers meet the expected state.
    * Failure: The test will fail if any of the NTP server does not meet the expected state.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyNTPAssociations:
          ntp_servers:
            - server_address: 1.1.1.1
              preferred: True
              stratum: 1
            - server_address: 2.2.2.2
              stratum: 2
            - server_address: 3.3.3.3
              stratum: 2
      - VerifyNTPAssociations:
          ntp_pool:
            server_addresses: [1.1.1.1, 2.2.2.2]
            preferred_stratum_range: [1,3]
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show ntp associations")]

    class Input(AntaTest.Input):
        """Input model for the VerifyNTPAssociations test."""

        ntp_servers: list[NTPServer] | None = None
        """List of NTP servers."""
        ntp_pool: NTPPool | None = None
        """NTP servers pool."""
        NTPServer: ClassVar[type[NTPServer]] = NTPServer

        @model_validator(mode="after")
        def validate_inputs(self) -> Self:
            """Validate the inputs provided to the VerifyNTPAssociations test.

            Either `ntp_servers` or `ntp_pool` can be provided at the same time.
            """
            if not self.ntp_servers and not self.ntp_pool:
                msg = "'ntp_servers' or 'ntp_pool' must be provided"
                raise ValueError(msg)
            if self.ntp_servers and self.ntp_pool:
                msg = "Either 'ntp_servers' or 'ntp_pool' can be provided at the same time"
                raise ValueError(msg)

            # Verifies the len of preferred_stratum_range in NTP Pool should be 2 as this is the range.
            stratum_range = 2
            if self.ntp_pool and len(self.ntp_pool.preferred_stratum_range) > stratum_range:
                msg = "'preferred_stratum_range' list should have at most 2 items"
                raise ValueError(msg)
            return self

    def _validate_ntp_server(self, ntp_server: NTPServer, peers: dict[str, Any]) -> list[str]:
        """Validate the NTP server, condition and stratum level."""
        failure_msgs: list[str] = []
        server_address = str(ntp_server.server_address)

        # We check `peerIpAddr` in the peer details - covering IPv4Address input, or the peer key - covering Hostname input.
        matching_peer = next((peer for peer, peer_details in peers.items() if (server_address in {peer_details["peerIpAddr"], peer})), None)

        if not matching_peer:
            failure_msgs.append(f"{ntp_server} - Not configured")
            return failure_msgs

        # Collecting the expected/actual NTP peer details.
        exp_condition = "sys.peer" if ntp_server.preferred else "candidate"
        exp_stratum = ntp_server.stratum
        act_condition = get_value(peers[matching_peer], "condition")
        act_stratum = get_value(peers[matching_peer], "stratumLevel")

        if act_condition != exp_condition:
            failure_msgs.append(f"{ntp_server} - Incorrect condition - Expected: {exp_condition} Actual: {act_condition}")

        if act_stratum != exp_stratum:
            failure_msgs.append(f"{ntp_server} - Incorrect stratum level - Expected: {exp_stratum} Actual: {act_stratum}")

        return failure_msgs

    def _validate_ntp_pool(self, server_addresses: list[Hostname | IPv4Address], peer: str, stratum_range: list[int], peer_details: dict[str, Any]) -> list[str]:
        """Validate the NTP server pool, condition and stratum level."""
        failure_msgs: list[str] = []

        # We check `peerIpAddr` and `peer` in the peer details - covering server_addresses input
        if (peer_ip := peer_details["peerIpAddr"]) not in server_addresses and peer not in server_addresses:
            failure_msgs.append(f"NTP Server: {peer_ip} Hostname: {peer} - Associated but not part of the provided NTP pool")
            return failure_msgs

        act_condition = get_value(peer_details, "condition")
        act_stratum = get_value(peer_details, "stratumLevel")

        if act_condition not in ["sys.peer", "candidate"]:
            failure_msgs.append(f"NTP Server: {peer_ip} Hostname: {peer} - Incorrect condition  - Expected: sys.peer, candidate Actual: {act_condition}")

        if int(act_stratum) not in range(stratum_range[0], stratum_range[1] + 1):
            msg = f"Expected Stratum Range: {stratum_range[0]} to {stratum_range[1]} Actual: {act_stratum}"
            failure_msgs.append(f"NTP Server: {peer_ip} Hostname: {peer} - Incorrect stratum level - {msg}")

        return failure_msgs

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyNTPAssociations."""
        self.result.is_success()

        if not (peers := get_value(self.instance_commands[0].json_output, "peers")):
            self.result.is_failure("No NTP peers configured")
            return

        if self.inputs.ntp_servers:
            # Iterate over each NTP server.
            for ntp_server in self.inputs.ntp_servers:
                failure_msgs = self._validate_ntp_server(ntp_server, peers)
                for msg in failure_msgs:
                    self.result.is_failure(msg)
            return

        # Verifies the NTP pool details
        server_addresses = self.inputs.ntp_pool.server_addresses
        exp_stratum_range = self.inputs.ntp_pool.preferred_stratum_range
        for peer, peer_details in peers.items():
            failure_msgs = self._validate_ntp_pool(server_addresses, peer, exp_stratum_range, peer_details)
            for msg in failure_msgs:
                self.result.is_failure(msg)


class VerifyMaintenance(AntaTest):
    """Verifies that the device is not currently under or entering maintenance.

    Expected Results
    ----------------
    * Success: The test will pass if the device is not under or entering maintenance.
    * Failure: The test will fail if the device is under or entering maintenance.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyMaintenance:
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show maintenance", revision=1)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyMaintenance."""
        self.result.is_success()

        # If units is not empty we have to examine the output for details.
        if not (units := get_value(self.instance_commands[0].json_output, "units")):
            return
        units_under_maintenance = [unit for unit, info in units.items() if info["state"] == "underMaintenance"]
        units_entering_maintenance = [unit for unit, info in units.items() if info["state"] == "maintenanceModeEnter"]
        causes = set()
        # Iterate over units, check for units under or entering maintenance, and examine the causes.
        for info in units.values():
            if info["adminState"] == "underMaintenance":
                causes.add("Quiesce is configured")
            if info["onBootMaintenance"]:
                causes.add("On-boot maintenance is configured")
            if info["intfsViolatingTrafficThreshold"]:
                causes.add("Interface traffic threshold violation")

        # Building the error message.
        if units_under_maintenance:
            self.result.is_failure(f"Units under maintenance: '{', '.join(units_under_maintenance)}'")
        if units_entering_maintenance:
            self.result.is_failure(f"Units entering maintenance: '{', '.join(units_entering_maintenance)}'")
        if causes:
            self.result.is_failure(f"Possible causes: '{', '.join(sorted(causes))}'")


class VerifyFlashUtilization(AntaTest):
    """Verifies the free space on the flash drive is sufficient.

    Expected Results
    ----------------
    * Success: The test will pass if flash utilization is below the defined threshold.
    * Failure: The test will fail if flash utilization is above the defined threshold.

    Examples
    --------
    ```yaml
    anta.tests.hardware:
      - VerifyFlashUtilization:
          max_utilization: 70
          check_peer_supervisor: True
    ```
    """

    categories: ClassVar[list[str]] = ["hardware"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaTemplate(template="{command}", revision=1)]

    class Input(AntaTest.Input):
        """Input model for the VerifyFlashUtilization test."""

        max_utilization: Percent = 70
        """The maximum allowed percentage of flash memory utilization."""
        check_peer_supervisor: bool = False
        """If True, also verifies the peer supervisor flash drive on dual-supervisor systems."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for peer supervisor."""
        if self.inputs.check_peer_supervisor:
            return [
                template.render(command="show file systems"),
                template.render(command="session peer-supervisor show file systems | json"),
            ]
        return [template.render(command="show file systems")]

    def _check_supervisor(self, output: dict[str, Any], supervisor_name: Literal["Active", "Standby"]) -> None:
        """Get and verify flash utilization for a single supervisor."""
        msg_prefix = f"{supervisor_name} Supervisor - " if self.inputs.check_peer_supervisor else ""

        if (drive_details := get_item(output["fileSystems"], "prefix", "flash:")) is None:
            self.result.is_failure(f"{msg_prefix}Flash not found")
            return

        free = int(drive_details["free"])
        size = int(drive_details["size"])

        if size == 0:
            self.result.is_failure(f"{msg_prefix}Flash reported a size of 0")
            return

        utilization = round(100 - (free / size) * 100, 2)

        if utilization > self.inputs.max_utilization:
            self.result.is_failure(f"{msg_prefix}Flash utilization above threshold - Expected: <= {self.inputs.max_utilization}% Actual: {utilization}%")

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyFlashUtilization."""
        self.result.is_success()
        self._check_supervisor(self.instance_commands[0].json_output, supervisor_name="Active")

        # If dual-supervisor systems
        if self.inputs.check_peer_supervisor:
            try:
                # This block will safely handle issues with the peer supervisor output
                peer_output_str = self.instance_commands[1].json_output["output"]
                output = loads(peer_output_str)
                self._check_supervisor(output, supervisor_name="Standby")
            except (KeyError, JSONDecodeError, TypeError) as exc:
                self.result.is_failure(f"Standby Supervisor - Failed to parse command output - {exc!s}")


class VerifyFilePresence(AntaTest):
    """Verifies the presence of files on the device flash memory.

    Expected Results
    ----------------
    * Success: The test will pass if all specified files are found on the flash drive.
    * Failure: The test will fail if any file is not found.

    Examples
    --------
    ```yaml
    anta.tests.system:
      - VerifyFilePresence:
          filenames:
            - script.py
          check_peer_supervisor: true
    ```
    """

    categories: ClassVar[list[str]] = ["system"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaTemplate(template="dir {flash_memory}", revision=2)]

    class Input(AntaTest.Input):
        """Input model for the VerifyFilePresence test."""

        filenames: list[str]
        """List of files to verify, including their extensions if any."""
        check_peer_supervisor: bool = False
        """If True, also verifies the peer supervisor flash drive on dual-supervisor systems."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template as per the input."""
        if self.inputs.check_peer_supervisor:
            return [template.render(flash_memory="flash:/"), template.render(flash_memory="supervisor-peer:/mnt/flash")]
        return [template.render(flash_memory="flash:/")]

    @skip_on_platforms(["cEOSLab", "vEOS-lab", "cEOSCloudLab", "vEOS"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyFilePresence."""
        self.result.is_success()

        # Verify the active supervisor
        directories = self.instance_commands[0].json_output
        for filename in self.inputs.filenames:
            if not get_value(directories, f"urls..{self.instance_commands[0].params.flash_memory}..entries..{filename}", separator=".."):
                self.result.is_failure(f"File: {filename} - Not found")

        # Verify the standby supervisor
        if self.inputs.check_peer_supervisor:
            peer_directories = self.instance_commands[1].json_output
            for filename in self.inputs.filenames:
                if not get_value(peer_directories, f"urls..{self.instance_commands[1].params.flash_memory}..entries..{filename}", separator=".."):
                    self.result.is_failure(f"File: {filename} - Not found on standby supervisor")



# Copyright (c) 2024-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Constants and Enums for the asynceapi package."""

from __future__ import annotations

from enum import Enum


class EapiCommandFormat(str, Enum):
    """Enum for the eAPI command format.

    NOTE: This could be updated to StrEnum when Python 3.11 is the minimum supported version in ANTA.
    """

    JSON = "json"
    TEXT = "text"

    def __str__(self) -> str:
        """Override the __str__ method to return the value of the Enum, mimicking the behavior of StrEnum."""
        return self.value



# Copyright (c) 2023-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Module related to the EOS various logging tests.

NOTE: The EOS command `show logging` does not support JSON output format.
"""

# Pyright does not understand AntaTest.Input typing
# pyright: reportAttributeAccessIssue=false
from __future__ import annotations

import re
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Annotated, ClassVar, Literal

from pydantic import Field

from anta.custom_types import LogSeverityLevel
from anta.input_models.logging import LoggingQuery
from anta.models import AntaCommand, AntaTemplate, AntaTest

if TYPE_CHECKING:
    import logging


def _get_logging_states(logger: logging.Logger, command_output: str) -> str:
    """Parse `show logging` output and gets operational logging states used in the tests in this module.

    Parameters
    ----------
    logger
        The logger object.
    command_output
        The `show logging` output.

    Returns
    -------
    str
        The operational logging states.

    """
    log_states = command_output.partition("\n\nExternal configuration:")[0]
    logger.debug("Device logging states:\n%s", log_states)
    return log_states


class VerifySyslogLogging(AntaTest):
    """Verifies if syslog logging is enabled.

    Expected Results
    ----------------
    * Success: The test will pass if syslog logging is enabled.
    * Failure: The test will fail if syslog logging is disabled.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifySyslogLogging:
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show logging", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifySyslogLogging."""
        self.result.is_success()
        log_output = self.instance_commands[0].text_output

        if "Syslog logging: enabled" not in _get_logging_states(self.logger, log_output):
            self.result.is_failure("Syslog logging is disabled")


class VerifyLoggingPersistent(AntaTest):
    """Verifies if logging persistent is enabled and logs are saved in flash.

    Expected Results
    ----------------
    * Success: The test will pass if logging persistent is enabled and logs are in flash.
    * Failure: The test will fail if logging persistent is disabled or no logs are saved in flash.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingPersistent:
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaCommand(command="show logging", ofmt="text"),
        AntaCommand(command="dir flash:/persist/messages", ofmt="text"),
    ]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingPersistent."""
        self.result.is_success()
        log_output = self.instance_commands[0].text_output
        dir_flash_output = self.instance_commands[1].text_output
        if "Persistent logging: disabled" in _get_logging_states(self.logger, log_output):
            self.result.is_failure("Persistent logging is disabled")
            return
        pattern = r"-rw-\s+(\d+)"
        persist_logs = re.search(pattern, dir_flash_output)
        if not persist_logs or int(persist_logs.group(1)) == 0:
            self.result.is_failure("No persistent logs are saved in flash")


class VerifyLoggingSourceIntf(AntaTest):
    """Verifies logging source-interface for a specified VRF.

    Expected Results
    ----------------
    * Success: The test will pass if the provided logging source-interface is configured in the specified VRF.
    * Failure: The test will fail if the provided logging source-interface is NOT configured in the specified VRF.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingSourceIntf:
          interface: Management0
          vrf: default
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show logging", ofmt="text")]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingSourceIntf test."""

        interface: str
        """Source-interface to use as source IP of log messages."""
        vrf: str = "default"
        """The name of the VRF to transport log messages. Defaults to `default`."""

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingSourceIntf."""
        output = self.instance_commands[0].text_output
        pattern = rf"Logging source-interface '{self.inputs.interface}'.*VRF {self.inputs.vrf}"
        if re.search(pattern, _get_logging_states(self.logger, output)):
            self.result.is_success()
        else:
            self.result.is_failure(f"Source-interface: {self.inputs.interface} VRF: {self.inputs.vrf} - Not configured")


class VerifyLoggingHosts(AntaTest):
    """Verifies logging hosts (syslog servers) for a specified VRF.

    Expected Results
    ----------------
    * Success: The test will pass if the provided syslog servers are configured in the specified VRF.
    * Failure: The test will fail if the provided syslog servers are NOT configured in the specified VRF.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingHosts:
          hosts:
            - 1.1.1.1
            - 2.2.2.2
          vrf: default
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show logging", ofmt="text")]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingHosts test."""

        hosts: list[IPv4Address]
        """List of hosts (syslog servers) IP addresses."""
        vrf: str = "default"
        """The name of the VRF to transport log messages. Defaults to `default`."""

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingHosts."""
        output = self.instance_commands[0].text_output
        not_configured = []
        for host in self.inputs.hosts:
            pattern = rf"Logging to '{host!s}'.*VRF {self.inputs.vrf}"
            if not re.search(pattern, _get_logging_states(self.logger, output)):
                not_configured.append(str(host))

        if not not_configured:
            self.result.is_success()
        else:
            self.result.is_failure(f"Syslog servers {', '.join(not_configured)} are not configured in VRF {self.inputs.vrf}")


class VerifyLoggingLogsGeneration(AntaTest):
    """Verifies if logs are generated.

    This test performs the following checks:

      1. Sends a test log message at the specified severity log level.
      2. Retrieves the most recent logs (last 30 seconds).
      3. Verifies that the test message was successfully logged.

    Expected Results
    ----------------
    * Success: If logs are being generated and the test message is found in recent logs.
    * Failure: If any of the following occur:
        - The test message is not found in recent logs.
        - The logging system is not capturing new messages.
        - No logs are being generated.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingLogsGeneration:
          severity_level: informational
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaTemplate(template="send log level {severity_level} message ANTA VerifyLoggingLogsGeneration validation", ofmt="text"),
        AntaTemplate(template="show logging {severity_level} last 30 seconds | grep ANTA", ofmt="text", use_cache=False),
    ]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingLogsGeneration test."""

        severity_level: LogSeverityLevel = "informational"
        """Log severity level. Defaults to informational."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for log severity level in the input."""
        return [template.render(severity_level=self.inputs.severity_level)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingLogsGeneration."""
        log_pattern = r"ANTA VerifyLoggingLogsGeneration validation"
        output = self.instance_commands[1].text_output
        lines = output.strip().split("\n")[::-1]
        for line in lines:
            if re.search(log_pattern, line):
                self.result.is_success()
                return
        self.result.is_failure("Logs are not generated")


class VerifyLoggingHostname(AntaTest):
    """Verifies if logs are generated with the device FQDN.

    This test performs the following checks:

      1. Retrieves the device's configured FQDN.
      2. Sends a test log message at the specified severity log level.
      3. Retrieves the most recent logs (last 30 seconds).
      4. Verifies that the test message includes the complete FQDN of the device.

    Expected Results
    ----------------
    * Success: If logs are generated with the device's complete FQDN.
    * Failure: If any of the following occur:
        - The test message is not found in recent logs.
        - The log message does not include the device's FQDN.
        - The FQDN in the log message doesn't match the configured FQDN.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingHostname:
          severity_level: informational
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaCommand(command="show hostname", revision=1),
        AntaTemplate(template="send log level {severity_level} message ANTA VerifyLoggingHostname validation", ofmt="text"),
        AntaTemplate(template="show logging {severity_level} last 30 seconds | grep ANTA", ofmt="text", use_cache=False),
    ]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingHostname test."""

        severity_level: LogSeverityLevel = "informational"
        """Log severity level. Defaults to informational."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for log severity level in the input."""
        return [template.render(severity_level=self.inputs.severity_level)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingHostname."""
        output_hostname = self.instance_commands[0].json_output
        output_logging = self.instance_commands[2].text_output
        fqdn = output_hostname["fqdn"]
        lines = output_logging.strip().split("\n")[::-1]
        log_pattern = r"ANTA VerifyLoggingHostname validation"
        last_line_with_pattern = ""
        for line in lines:
            if re.search(log_pattern, line):
                last_line_with_pattern = line
                break
        if fqdn in last_line_with_pattern:
            self.result.is_success()
        else:
            self.result.is_failure("Logs are not generated with the device FQDN")


class VerifyLoggingTimestamp(AntaTest):
    """Verifies if logs are generated with the appropriate timestamp.

    This test performs the following checks:

      1. Sends a test log message at the specified severity log level.
      2. Retrieves the most recent logs (last 30 seconds).
      3. Verifies that the test message is present with a high-resolution RFC3339 timestamp format.
        - Example format: `2024-01-25T15:30:45.123456+00:00`.
        - Includes microsecond precision.
        - Contains timezone offset.

    Expected Results
    ----------------
    * Success: If logs are generated with the correct high-resolution RFC3339 timestamp format.
    * Failure: If any of the following occur:
        - The test message is not found in recent logs.
        - The timestamp format does not match the expected RFC3339 format.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingTimestamp:
          severity_level: informational
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [
        AntaTemplate(template="send log level {severity_level} message ANTA VerifyLoggingTimestamp validation", ofmt="text"),
        AntaTemplate(template="show logging {severity_level} last 30 seconds | grep ANTA", ofmt="text", use_cache=False),
    ]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingTimestamp test."""

        severity_level: LogSeverityLevel = "informational"
        """Log severity level. Defaults to informational."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for log severity level in the input."""
        return [template.render(severity_level=self.inputs.severity_level)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingTimestamp."""
        log_pattern = r"ANTA VerifyLoggingTimestamp validation"
        timestamp_pattern = r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}[+-]\d{2}:\d{2}"
        output = self.instance_commands[1].text_output
        lines = output.strip().split("\n")[::-1]
        last_line_with_pattern = ""
        for line in lines:
            if re.search(log_pattern, line):
                last_line_with_pattern = line
                break
        if re.search(timestamp_pattern, last_line_with_pattern):
            self.result.is_success()
        else:
            self.result.is_failure("Logs are not generated with the appropriate timestamp format")


class VerifyLoggingAccounting(AntaTest):
    """Verifies if AAA accounting logs are generated.

    Expected Results
    ----------------
    * Success: The test will pass if AAA accounting logs are generated.
    * Failure: The test will fail if AAA accounting logs are NOT generated.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingAccounting:
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaCommand(command="show aaa accounting logs | tail", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingAccounting."""
        pattern = r"cmd=show aaa accounting logs"
        output = self.instance_commands[0].text_output
        if re.search(pattern, output):
            self.result.is_success()
        else:
            self.result.is_failure("AAA accounting logs are not generated")


class VerifyLoggingErrors(AntaTest):
    """Verifies there are no syslog messages with a severity of ERRORS or higher.

    Expected Results
    ----------------
      * Success: The test will pass if there are NO syslog messages with a severity of ERRORS or higher.
      * Failure: The test will fail if ERRORS or higher syslog messages are present.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingErrors:
          last_number_time_units: 10
          time_unit: hours
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaTemplate(template="show logging threshold errors {log_history_depth}", ofmt="text")]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingErrors test."""

        last_number_time_units: Annotated[int, Field(ge=1, le=9999)] | None = None
        """Number of time units to look in the logging buffers.

        The actual duration is determined based on the selected `time_unit` (e.g. 5 days, 10 minutes)."""
        time_unit: Literal["days", "hours", "minutes", "seconds"] = "days"
        """Unit of time to be used with `last_number_time_units`."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for log history depth in the input."""
        commands: list[AntaCommand] = []
        log_history_depth: str = ""

        if self.inputs.last_number_time_units:
            log_history_depth = f"last {self.inputs.last_number_time_units} {self.inputs.time_unit}"
        commands.append(template.render(log_history_depth=log_history_depth))

        return commands

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingErrors."""
        command_output = self.instance_commands[0].text_output

        if len(command_output) == 0:
            self.result.is_success()
        else:
            self.result.is_failure(f"Device has reported syslog messages with a severity of ERRORS or higher:\n{command_output}")


class VerifyLoggingEntries(AntaTest):
    """Verifies that log strings are present or absent in the logging buffer.

    Expected Results
    ----------------
    * Success:
        - If `fail_on_match` is `False` (default): the test will pass if all specified patterns are found in the log output.
        - If `fail_on_match` is `True`: the test will pass if none of the specified patterns are found in the log output.

    * Failure:
        - If `fail_on_match` is `False`: the test will fail if any of the specified patterns are missing from the log output.
        - If `fail_on_match` is `True`: the test will fail if any of the specified patterns are found in the log output.

    Examples
    --------
    ```yaml
    anta.tests.logging:
      - VerifyLoggingEntries:
          logging_entries:
            - regex_match: ".*ACCOUNTING-5-EXEC: cvpadmin ssh.*"
              last_number_messages: 30
              severity_level: alerts
            - regex_match:
                - ".*SPANTREE-6-INTERFACE_ADD:.*"
                - ".*ACCOUNTING-5-EXEC: cvpadmin ssh.*"
              last_number_time_units: 3
              time_unit: hours
              severity_level: critical
              fail_on_match: True
    ```
    """

    categories: ClassVar[list[str]] = ["logging"]
    commands: ClassVar[list[AntaCommand | AntaTemplate]] = [AntaTemplate(template="show logging {log_history_depth} {severity_level}", ofmt="text", use_cache=False)]

    class Input(AntaTest.Input):
        """Input model for the VerifyLoggingEntries test."""

        logging_entries: list[LoggingQuery]
        """List of logging entries and regex match."""

    def render(self, template: AntaTemplate) -> list[AntaCommand]:
        """Render the template for log history depth and log severity level in the input."""
        commands: list[AntaCommand] = []
        for entry in self.inputs.logging_entries:
            log_history_depth: str | int = ""
            if entry.last_number_messages:
                log_history_depth = entry.last_number_messages
            elif entry.last_number_time_units:
                log_history_depth = f"last {entry.last_number_time_units} {entry.time_unit}"
            commands.append(template.render(log_history_depth=log_history_depth, severity_level=entry.severity_level))
        return commands

    @AntaTest.anta_test
    def test(self) -> None:
        """Main test function for VerifyLoggingEntries."""
        self.result.is_success()
        for command_output, logging_entry in zip(self.instance_commands, self.inputs.logging_entries, strict=False):
            output = command_output.text_output
            log_history_depth = command_output.params.log_history_depth
            patterns_to_check = logging_entry.regex_match if isinstance(logging_entry.regex_match, list) else [logging_entry.regex_match]

            for pattern in patterns_to_check:
                match_found = re.search(pattern, output)

                # Fails if any of the regex patterns are found
                if logging_entry.fail_on_match and match_found:
                    self.result.is_failure(f"Unexpected Pattern: `{pattern}` - Found in last {log_history_depth} {logging_entry.severity_level} log entries")

                # Fails if any of the regex patterns not found
                if not logging_entry.fail_on_match and not match_found:
                    self.result.is_failure(f"Pattern: `{pattern}` - Not found in last {log_history_depth} {logging_entry.severity_level} log entries")



# Copyright (c) 2024-2025 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the LICENSE file.
"""Models for the asynceapi package."""

from __future__ import annotations

from dataclasses import dataclass, field
from json import JSONDecodeError, loads
from logging import getLogger
from typing import TYPE_CHECKING, Any, Literal
from uuid import uuid4

from ._constants import EapiCommandFormat
from ._errors import EapiReponseError

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ._types import EapiComplexCommand, EapiJsonOutput, EapiSimpleCommand, EapiTextOutput, JsonRpc

LOGGER = getLogger(__name__)


@dataclass(frozen=True)
class EapiRequest:
    """Model for an eAPI request.

    Attributes
    ----------
    commands : list[EapiSimpleCommand | EapiComplexCommand]
        A list of commands to execute.
    version : int | Literal["latest"]
        The eAPI version to use. Defaults to "latest".
    format : EapiCommandFormat
        The command output format. Defaults "json".
    timestamps : bool
        Include timestamps in the command output. Defaults to False.
    auto_complete : bool
        Enable command auto-completion. Defaults to False.
    expand_aliases : bool
        Expand command aliases. Defaults to False.
    stop_on_error : bool
        Stop command execution on first error. Defaults to True.
    id : int | str
        The request ID. Defaults to a random hex string.
    """

    commands: list[EapiSimpleCommand | EapiComplexCommand]
    version: int | Literal["latest"] = "latest"
    format: EapiCommandFormat = EapiCommandFormat.JSON
    timestamps: bool = False
    auto_complete: bool = False
    expand_aliases: bool = False
    stop_on_error: bool = True
    id: int | str = field(default_factory=lambda: uuid4().hex)

    def to_jsonrpc(self) -> JsonRpc:
        """Return the JSON-RPC dictionary payload for the request."""
        return {
            "jsonrpc": "2.0",
            "method": "runCmds",
            "params": {
                "version": self.version,
                "cmds": self.commands,
                "format": self.format,
                "timestamps": self.timestamps,
                "autoComplete": self.auto_complete,
                "expandAliases": self.expand_aliases,
                "stopOnError": self.stop_on_error,
            },
            "id": self.id,
        }


@dataclass(frozen=True)
class EapiResponse:
    """Model for an eAPI response.

    Construct an EapiResponse from a JSON-RPC response dictionary using the `from_jsonrpc` class method.

    Can be iterated over to access command results in order of execution.

    Attributes
    ----------
    request_id : str
        The ID of the original request this response corresponds to.
    _results : dict[int, EapiCommandResult]
        Dictionary mapping request command indices to their respective results.
    error_code : int | None
        The JSON-RPC error code, if any.
    error_message : str | None
        The JSON-RPC error message, if any.
    """

    request_id: str
    _results: dict[int, EapiCommandResult] = field(default_factory=dict)
    error_code: int | None = None
    error_message: str | None = None

    @property
    def success(self) -> bool:
        """Return True if the response has no errors."""
        return self.error_code is None

    @property
    def results(self) -> list[EapiCommandResult]:
        """Get all results as a list. Results are ordered by the command indices in the request."""
        return list(self._results.values())

    def __len__(self) -> int:
        """Return the number of results."""
        return len(self._results)

    def __iter__(self) -> Iterator[EapiCommandResult]:
        """Enable iteration over the results. Results are yielded in the same order as provided in the request."""
        yield from self._results.values()

    @classmethod
    def from_jsonrpc(cls, response: dict[str, Any], request: EapiRequest, *, raise_on_error: bool = False) -> EapiResponse:
        """Build an EapiResponse from a JSON-RPC eAPI response.

        Parameters
        ----------
        response
            The JSON-RPC eAPI response dictionary.
        request
            The corresponding EapiRequest.
        raise_on_error
            Raise an EapiReponseError if the response contains errors, by default False.

        Returns
        -------
        EapiResponse
            The EapiResponse object.
        """
        has_error = "error" in response
        response_data = response["error"]["data"] if has_error else response["result"]

        # Handle case where we have fewer results than commands (stop_on_error=True)
        executed_count = min(len(response_data), len(request.commands))

        # Process the results we have
        results = {}
        for i in range(executed_count):
            cmd = request.commands[i]
            cmd_str = cmd["cmd"] if isinstance(cmd, dict) else cmd
            data = response_data[i]

            output = None
            errors = []
            success = True
            start_time = None
            duration = None

            # Parse the output based on the data type, no output when errors are present
            if isinstance(data, dict):
                if "errors" in data:
                    errors = data["errors"]
                    success = False
                else:
                    output = data["output"] if request.format == EapiCommandFormat.TEXT and "output" in data else data

                # Add timestamps if available
                if request.timestamps and "_meta" in data:
                    meta = data.pop("_meta")
                    start_time = meta.get("execStartTime")
                    duration = meta.get("execDuration")

            elif isinstance(data, str):
                # Handle case where eAPI returns a JSON string response (serialized JSON) for certain commands
                try:
                    output = loads(data)
                except (JSONDecodeError, TypeError):
                    # If it's not valid JSON, store as is
                    LOGGER.warning("Invalid JSON response for command: %s. Storing as text: %s", cmd_str, data)
                    output = data

            results[i] = EapiCommandResult(
                command=cmd_str,
                output=output,
                errors=errors,
                success=success,
                start_time=start_time,
                duration=duration,
            )

        # If stop_on_error is True and we have an error, indicate commands not executed
        if has_error and request.stop_on_error and executed_count < len(request.commands):
            for i in range(executed_count, len(request.commands)):
                cmd = request.commands[i]
                cmd_str = cmd["cmd"] if isinstance(cmd, dict) else cmd
                results[i] = EapiCommandResult(command=cmd_str, output=None, errors=["Command not executed due to previous error"], success=False, executed=False)

        response_obj = cls(
            request_id=response["id"],
            _results=results,
            error_code=response["error"]["code"] if has_error else None,
            error_message=response["error"]["message"] if has_error else None,
        )

        if raise_on_error and has_error:
            raise EapiReponseError(response_obj)

        return response_obj


@dataclass(frozen=True)
class EapiCommandResult:
    """Model for an eAPI command result.

    Attributes
    ----------
    command : str
        The command that was executed.
    output : EapiJsonOutput | EapiTextOutput | None
        The command result output. None if the command returned errors.
    errors : list[str]
        A list of error messages, if any.
    success : bool
        True if the command was successful.
    executed : bool
        True if the command was executed. When `stop_on_error` is True in the request, some commands may not be executed.
    start_time : float | None
        Command execution start time in seconds. Uses Unix epoch format. `timestamps` must be True in the request.
    duration : float | None
        Command execution duration in seconds. `timestamps` must be True in the request.
    """

    command: str
    output: EapiJsonOutput | EapiTextOutput | None
    errors: list[str] = field(default_factory=list)
    success: bool = True
    executed: bool = True
    start_time: float | None = None
    duration: float | None = None
